from ....units import Units


class ArrayMixin:
    """Mixin class for a container of an array.

    .. versionadded:: TODODASKVER

    """

    def __array_function__(self, func, types, args, kwargs):
        """Implement the `numpy` ``__array_function__`` protocol.

        .. versionadded:: TODODASKVER

        """
        return NotImplemented

    @property
    def Units(self):
        """The `cf.Units` object containing the units of the array.

        .. versionadded:: TODODASKVER

        """
        return Units(self.get_units(), self.get_calendar(None))



from math import prod

import cfdm

from . import mixin
from .constructs import Constructs
from .data import Data
from .decorators import _inplace_enabled, _inplace_enabled_define_and_cleanup
from .functions import (
    _DEPRECATION_ERROR_ARG,
    _DEPRECATION_ERROR_METHOD,
    indices_shape,
    parse_indices,
)

_empty_set = set()


class Domain(mixin.FieldDomain, mixin.Properties, cfdm.Domain):
    """A domain construct of the CF data model.

    The domain represents a set of discrete "locations" in what
    generally would be a multi-dimensional space, either in the real
    world or in a model's simulated world. The data array elements of
    a field construct correspond to individual location of a domain.

    The domain construct is defined collectively by the following
    constructs of the CF data model: domain axis, dimension
    coordinate, auxiliary coordinate, cell measure, coordinate
    reference, and domain ancillary constructs; as well as properties
    to describe the domain.

    **NetCDF interface**

    {{netCDF variable}}

    {{netCDF global attributes}}

    {{netCDF group attributes}}

    {{netCDF geometry group}}

    Some components exist within multiple constructs, but when written
    to a netCDF dataset the netCDF names associated with such
    components will be arbitrarily taken from one of them. The netCDF
    variable, dimension and sample dimension names and group
    structures for such components may be set or removed consistently
    across all such components with the `nc_del_component_variable`,
    `nc_set_component_variable`, `nc_set_component_variable_groups`,
    `nc_clear_component_variable_groups`,
    `nc_del_component_dimension`, `nc_set_component_dimension`,
    `nc_set_component_dimension_groups`,
    `nc_clear_component_dimension_groups`,
    `nc_del_component_sample_dimension`,
    `nc_set_component_sample_dimension`,
    `nc_set_component_sample_dimension_groups`,
    `nc_clear_component_sample_dimension_groups` methods.

    A domain construct of the CF data model.

    The domain represents a set of discrete "locations" in what
    generally would be a multi-dimensional space, either in the real
    world or in a model's simulated world. The data array elements of
    a field construct correspond to individual location of a domain.

    The domain construct is defined collectively by the following
    constructs of the CF data model: domain axis, dimension
    coordinate, auxiliary coordinate, cell measure, coordinate
    reference, and domain ancillary constructs; as well as properties
    to describe the domain.

    """

    def __new__(cls, *args, **kwargs):
        """Creates a new Domain instance."""
        instance = super().__new__(cls)
        instance._Data = Data
        instance._Constructs = Constructs
        return instance

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        return super().__repr__().replace("<", "<CF ", 1)

    @property
    def _cyclic(self):
        """Storage for axis cyclicity.

        Do not change the value in-place.

        """
        return self._custom.get("_cyclic", _empty_set)

    @_cyclic.setter
    def _cyclic(self, value):
        """value must be a set.

        Do not change the value in-place.

        """
        self._custom["_cyclic"] = value

    @_cyclic.deleter
    def _cyclic(self):
        self._custom["_cyclic"] = _empty_set

    @property
    def size(self):
        """The number of locations in the domain.

        If there are no domain axis constructs, or any domain axis
        construct has a size of 0, then the size is 0.

        """
        domain_axes = self.domain_axes(todict=True)
        if not domain_axes:
            return 0

        return prod(
            [domain_axis.get_size(0) for domain_axis in domain_axes.values()]
        )

    def close(self):
        """Close all files referenced by the domain construct.

        Deprecated at version TODODASKVER. All files are now
        automatically closed when not being accessed.

        Note that a closed file will be automatically reopened if its
        contents are subsequently required.

        :Returns:

            `None`

        **Examples**

        >>> d.close()

        """
        _DEPRECATION_ERROR_METHOD(
            self,
            "close",
            "All files are now automatically closed when not being accessed.",
            version="TODODASKVER",
            removed_at="5.0.0",
        )  # pragma: no cover

    @_inplace_enabled(default=False)
    def flip(self, axes=None, inplace=False):
        """Flip (reverse the direction of) domain axes.

        .. seealso:: `domain_axis`, `transpose`

        :Parameters:

            axes: (sequence of) `str` , optional
                Select the domain axes to flip.

                A domain axis is identified by that which would be
                selected by passing a given axis description to a call of
                the `domain_axis` method. For example, a value of ``'X'``
                would select the domain axis construct returned by
                ``f.domain_axis('X')``.

                If no axes are provided then all axes are flipped.

            {{inplace: `bool`, optional}}

        :Returns:

            `Domain` or `None`
                The domain with flipped axes, or `None` if the operation
                was in-place.

        **Examples**

        >>> d = cf.example_field(0).domain
        >>> print(d)
        Dimension coords: latitude(5) = [-75.0, ..., 75.0] degrees_north
                        : longitude(8) = [22.5, ..., 337.5] degrees_east
                        : time(1) = [2019-01-01 00:00:00]

        >>> print(d.flip('X'))
        Dimension coords: latitude(5) = [-75.0, ..., 75.0] degrees_north
                        : longitude(8) = [337.5, ..., 22.5] degrees_east
                        : time(1) = [2019-01-01 00:00:00]

        >>> print(d.flip(['T', 'Y']))
        Dimension coords: latitude(5) = [75.0, ..., -75.0] degrees_north
                        : longitude(8) = [22.5, ..., 337.5] degrees_east
                        : time(1) = [2019-01-01 00:00:00]

        >>> print(d.flip())
        Dimension coords: latitude(5) = [75.0, ..., -75.0] degrees_north
                        : longitude(8) = [337.5, ..., 22.5] degrees_east
                        : time(1) = [2019-01-01 00:00:00]

        """
        d = _inplace_enabled_define_and_cleanup(self)

        if axes is None:
            # Flip all the axes
            axes = self.domain_axes(todict=True)
        else:
            axes = self._parse_axes(axes)

        axes = set(axes)

        # Flip constructs with data
        d.constructs._flip(axes)

        return d

    def get_data(self, default=ValueError(), _units=None, _fill_value=True):
        """Return a default value when data is requested.

        A `Domain` instance can never have data, so a default value
        must be returned if data is requested. This is useful for
        cases when it is not known in advance if a `Field` or `Domain`
        instance is in use.

        .. versionadded:: 3.11.0

        .. seealso:: `has_data`

        :Parameters:

            default: optional
                Return the value of the *default* parameter.

                {{default Exception}}

            _units: optional
                Ignored.

            _fill_value: optional
                Ignored.

        :Returns:

                The value of the *default* parameter, if an exception
                has not been raised.

        **Examples**

        >>> d = cf.example_domain(0)
        >>> print(d.get_data(None))
        None
        >>> d.get_data()
        Traceback (most recent call last):
            ...
        ValueError: Domain has no data

        """
        if default is None:
            return

        return self._default(
            default, message=f"{self.__class__.__name__} has no data"
        )

    def identity(self, default="", strict=False, relaxed=False, nc_only=False):
        """Return the canonical identity.

        By default the identity is the first found of the following:

        * The "id" attribute, preceded by ``'id%'``.
        * The "cf_role" property, preceded by ``'cf_role='``.
        * The "long_name" property, preceded by ``'long_name='``.
        * The netCDF variable name, preceded by ``'ncvar%'``.
        * The value of the *default* parameter.

        .. versionadded:: 3.11.0

        .. seealso:: `id`, `identities`

        :Parameters:

            default: optional
                If no identity can be found then return the value of the
                default parameter.

            strict: `bool`, optional
                If True then the identity is the first found of only the
                "standard_name" property or the "id" attribute.

            relaxed: `bool`, optional
                If True then the identity is the first found of only the
                "standard_name" property, the "id" attribute, the
                "long_name" property or the netCDF variable name.

            nc_only: `bool`, optional
                If True then only take the identity from the netCDF
                variable name.

        :Returns:

                The identity.

        **Examples**

        >>> f.properties()
        {'foo': 'bar',
         'long_name': 'Air Temperature',
         'standard_name': 'air_temperature'}
        >>> f.nc_get_variable()
        'tas'
        >>> f.identity()
        'air_temperature'
        >>> f.del_property('standard_name')
        'air_temperature'
        >>> f.identity(default='no identity')
        'air_temperature'
        >>> f.identity()
        'long_name=Air Temperature'
        >>> f.del_property('long_name')
        >>> f.identity()
        'ncvar%tas'
        >>> f.nc_del_variable()
        'tas'
        >>> f.identity()
        'ncvar%tas'
        >>> f.identity()
        ''
        >>> f.identity(default='no identity')
        'no identity'

        """
        if nc_only:
            if strict:
                raise ValueError(
                    "'strict' and 'nc_only' parameters cannot both be True"
                )

            if relaxed:
                raise ValueError(
                    "'relaxed' and 'nc_only' parameters cannot both be True"
                )

            n = self.nc_get_variable(None)
            if n is not None:
                return f"ncvar%{n}"

            return default

        n = getattr(self, "id", None)
        if n is not None:
            return f"id%{n}"

        if relaxed:
            n = self.get_property("long_name", None)
            if n is not None:
                return f"long_name={n}"

            n = self.nc_get_variable(None)
            if n is not None:
                return f"ncvar%{n}"

            return default

        if strict:
            return default

        for prop in ("cf_role", "long_name"):
            n = self.get_property(prop, None)
            if n is not None:
                return f"{prop}={n}"

        n = self.nc_get_variable(None)
        if n is not None:
            return f"ncvar%{n}"

        return default

    def identities(self):
        """Return all possible identities.

        The identities comprise:

        * The "id" attribute, preceded by ``'id%'``.
        * The ``cf_role`` property, preceded by ``'cf_role='``.
        * The ``long_name`` property, preceded by ``'long_name='``.
        * All other properties, preceded by the property name and a
          equals e.g. ``'foo=bar'``.
        * The netCDF variable name, preceded by ``'ncvar%'``.

        .. versionadded:: (cfdm) 1.9.0.0

        .. seealso:: `identity`

        :Returns:

            `list`
                The identities.

        **Examples**

        >>> d = {{package}}.Domain()
        >>> d.set_properties({'foo': 'bar',
        ...                   'long_name': 'Domain for model'})
        >>> d.nc_set_variable('dom1')
        >>> d.identities()
        ['long_name=Domain for model', 'foo=bar', 'ncvar%dom1']

        """
        out = super().identities()

        i = getattr(self, "id", None)
        if i is not None:
            # Insert id attribute
            i = f"id%{i}"
            if not out:
                out = [i]
            else:
                out.insert(0, i)

        return out

    def indices(self, *mode, **kwargs):
        """Create indices that define a subspace of the domain
        construct.

        The indices returned by this method be used to create the
        subspace by passing them to the `subspace` method of the
        original domain construct.

        The subspace is defined by identifying indices based on the
        metadata constructs.

        Metadata constructs are selected conditions are specified on
        their data. Indices for subspacing are then automatically
        inferred from where the conditions are met.

        Metadata constructs and the conditions on their data are
        defined by keyword parameters.

        * Any domain axes that have not been identified remain
          unchanged.

        * Multiple domain axes may be subspaced simultaneously, and it
          doesn't matter which order they are specified in.

        * Explicit indices may also be assigned to a domain axis
          identified by a metadata construct, with either a Python
          `slice` object, or a sequence of integers or booleans.

        * For a dimension that is cyclic, a subspace defined by a
          slice or by a `Query` instance is assumed to "wrap" around
          the edges of the data.

        * Conditions may also be applied to multi-dimensional metadata
          constructs. The "compress" mode is still the default mode
          (see the positional arguments), but because the indices may
          not be acting along orthogonal dimensions, some missing data
          may still need to be inserted into the field construct's
          data.

        .. versionadded:: 3.11.0

        .. seealso:: `subspace`, `where`, `__getitem__`,
                     `__setitem__`, `cf.Field.indices`

        :Parameters:

            mode: `str`, *optional*
                There are two modes of operation, each of which provides
                indices for a different type of subspace:

                ==============  ======================================
                *mode*          Description
                ==============  ======================================
                ``'compress'``  This is the default mode. Unselected
                                locations are removed to create the
                                returned subspace. Note that if a
                                multi-dimensional metadata construct
                                is being used to define the indices
                                then some missing data may still be
                                inserted at unselected locations.

                ``'envelope'``  The returned subspace is the smallest
                                that contains all of the selected
                                indices.
                ==============  ======================================

            kwargs: *optional*
                A keyword name is an identity of a metadata construct,
                and the keyword value provides a condition for
                inferring indices that apply to the dimension (or
                dimensions) spanned by the metadata construct's
                data. Indices are created that select every location
                for which the metadata construct's data satisfies the
                condition.

        :Returns:

            `dict`
                A dictionary of indices, keyed by the domain axis
                construct identifiers to which they apply.

        **Examples**

        >>> d = cf.example_field(0).domain
        >>> print(d)
        Dimension coords: latitude(5) = [-75.0, ..., 75.0] degrees_north
                        : longitude(8) = [22.5, ..., 337.5] degrees_east
                        : time(1) = [2019-01-01 00:00:00]
        >>> indices = d.indices(X=112.5)
        >>> indices
        {'domainaxis0': slice(0, 5, 1),
         'domainaxis1': slice(2, 3, 1),
         'domainaxis2': slice(0, 1, 1)}
        >>> print(d.subspace(**indices))
        Dimension coords: latitude(5) = [-75.0, ..., 75.0] degrees_north
                        : longitude(1) = [112.5] degrees_east
                        : time(1) = [2019-01-01 00:00:00]

        >>> indices = d.indices(X=112.5, Y=cf.wi(-60, 30))
        >>> indices
        {'domainaxis0': slice(1, 3, 1),
         'domainaxis1': slice(2, 3, 1),
         'domainaxis2': slice(0, 1, 1)}
        >>> print(d.subspace(**indices))
        Dimension coords: latitude(2) = [-45.0, 0.0] degrees_north
                        : longitude(1) = [112.5] degrees_east
                        : time(1) = [2019-01-01 00:00:00]

        >>> d.indices(X=[-1, 0], Y=slice(1, -1))
        {'domainaxis0': slice(1, 4, 1),
         'domainaxis1': slice(7, None, -7),
         'domainaxis2': slice(0, 1, 1)}
        >>> print(print(d.subspace(**indices)))
        Dimension coords: latitude(3) = [-45.0, 0.0, 45.0] degrees_north
                        : longitude(2) = [337.5, 22.5] degrees_east
                        : time(1) = [2019-01-01 00:00:00]

        """
        if len(mode) > 1:
            raise ValueError(
                "Can't provide more than one positional argument. "
                f"Got: {', '.join(repr(x) for x in mode)}"
            )

        if not mode or "compress" in mode:
            mode = "compress"
        elif "envelope" in mode:
            mode = "envelope"
        else:
            raise ValueError(f"Invalid value for 'mode' argument: {mode[0]!r}")

        # Get the indices for every domain axis in the domain, without
        # any auxiliary masks.
        domain_indices = self._indices(mode, None, False, kwargs)

        return domain_indices["indices"]

    def match_by_construct(self, *identities, OR=False, **conditions):
        """Whether or not there are particular metadata constructs.

        .. versionadded:: 3.11.0

        .. seealso:: `match`, `match_by_property`, `match_by_rank`,
                     `match_by_identity`, `match_by_ncvar`

        :Parameters:

            identities: optional
                Identify the metadata constructs by one or more of

                * A metadata construct identity.

                  {{construct selection identity}}

                * The key of a metadata construct

                If a cell method construct identity is given (such as
                ``'method:mean'``) then it will only be compared with the
                most recently applied cell method operation.

                Alternatively, one or more cell method constructs may
                be identified in a single string with a CF-netCDF cell
                methods-like syntax for describing both the collapse
                dimensions, the collapse method, and any cell method
                construct qualifiers. If N cell methods are described
                in this way then they will collectively identify the N
                most recently applied cell method operations. For
                example, ``'T: maximum within years T: mean over
                years'`` will be compared with the most two most
                recently applied cell method operations.

                *Parameter example:*
                  ``identity='latitude'``

                *Parameter example:*
                  ``'T'

                *Parameter example:*
                  ``'latitude'``

                *Parameter example:*
                  ``'long_name=Cell Area'``

                *Parameter example:*
                  ``'cellmeasure1'``

                *Parameter example:*
                  ``'measure:area'``

                *Parameter example:*
                  ``cf.eq('time')'``

                *Parameter example:*
                  ``re.compile('^lat')``

                *Parameter example:*
                  ``'domainancillary2', 'longitude'``

                *Parameter example:*
                  ``'area: mean T: maximum'``

                *Parameter example:*
                  ``'grid_latitude', 'area: mean T: maximum'``

            conditions: optional
                Identify the metadata constructs that have any of the
                given identities or construct keys, and whose data satisfy
                conditions.

                A construct identity or construct key (as defined by the
                *identities* parameter) is given as a keyword name and a
                condition on its data is given as the keyword value.

                The condition is satisfied if any of its data values
                equals the value provided.

                *Parameter example:*
                  ``longitude=180.0``

                *Parameter example:*
                  ``time=cf.dt('1959-12-16')``

                *Parameter example:*
                  ``latitude=cf.ge(0)``

                *Parameter example:*
                  ``latitude=cf.ge(0), air_pressure=500``

                *Parameter example:*
                  ``**{'latitude': cf.ge(0), 'long_name=soil_level': 4}``

            OR: `bool`, optional
                If True then return `True` if at least one metadata
                construct matches at least one of the criteria given by
                the *identities* or *conditions* arguments. By default
                `True` is only returned if the field constructs matches
                each of the given criteria.

        :Returns:

            `bool`
                Whether or not the domain construct contains the
                specified metadata constructs.

        **Examples**

        >>> d = cf.example_field(0).domain
        >>> print(d)
        Dimension coords: latitude(5) = [-75.0, ..., 75.0] degrees_north
                        : longitude(8) = [22.5, ..., 337.5] degrees_east
                        : time(1) = [2019-01-01 00:00:00]
        >>> d.match_by_construct("latitude")
        True
        >>> d.match_by_construct("air_pressure")
        False
        >>> d.match_by_construct("longitude", "time")
        True
        >>> d.match_by_construct(longitude=22.5)
        True
        >>> d.match_by_construct(longitude=15.5)
        False
        >>> d.match_by_construct(longitude=cf.gt(340))
        False
        >>> d.match_by_construct(longitude=cf.gt(240))
        True
        >>> d.match_by_construct(time=cf.dt("2019-01-01"))
        True
        >>> d.match_by_construct(time=cf.dt("2020-01-01"))
        False

        """
        if identities:
            if identities[0] == "or":
                _DEPRECATION_ERROR_ARG(
                    self,
                    "match_by_construct",
                    "or",
                    message="Use 'OR=True' instead.",
                    version="3.1.0",
                )  # pragma: no cover

            if identities[0] == "and":
                _DEPRECATION_ERROR_ARG(
                    self,
                    "match_by_construct",
                    "and",
                    message="Use 'OR=False' instead.",
                    version="3.1.0",
                )  # pragma: no cover

        if not identities and not conditions:
            return True

        constructs = self.constructs

        if not constructs:
            return False

        n = 0

        for identity in identities:
            filtered = constructs(identity)
            if filtered:
                n += 1
            elif not OR:
                return False

        if conditions:
            for identity, value in conditions.items():
                if self.subspace("test", **{identity: value}):
                    n += 1
                elif not OR:
                    return False

        if OR:
            return bool(n)

        return True

    @_inplace_enabled(default=False)
    def roll(self, axis, shift, inplace=False):
        """Roll the field along a cyclic axis.

        A unique axis is selected with the axes and kwargs parameters.

        .. versionadded:: 1.0

        .. seealso:: `anchor`, `axis`, `cyclic`, `iscyclic`, `period`

        :Parameters:

            axis:
                The cyclic axis to be rolled, defined by that which would
                be selected by passing the given axis description to a
                call of the field construct's `domain_axis` method. For
                example, for a value of ``'X'``, the domain axis construct
                returned by ``f.domain_axis('X')`` is selected.

            shift: `int`
                The number of places by which the selected cyclic axis is
                to be rolled.

            {{inplace: `bool`, optional}}

        :Returns:

            `Field`
                The rolled field.

        **Examples**

        Roll the data of the "X" axis one elements to the right:

        >>> f.roll('X', 1)

        Roll the data of the "X" axis three elements to the left:

        >>> f.roll('X', -3)

        """
        # TODODASK - allow multiple roll axes

        axis = self.domain_axis(
            axis,
            key=True,
            default=ValueError(
                f"Can't roll {self.__class__.__name__}. "
                f"Bad axis specification: {axis!r}"
            ),
        )

        d = _inplace_enabled_define_and_cleanup(self)

        # Roll the metadata constructs in-place
        axes = d._parse_axes(axis)
        d._roll_constructs(axes, shift)

        return d

    def subspace(self, *mode, **kwargs):
        """Create indices that define a subspace of the domain
        construct.

        The indices returned by this method be used to create the subspace
        by passing them to the `subspace` method of the original domain
        construct.

        The subspace is defined by identifying indices based on the
        metadata constructs.

        Metadata constructs are selected conditions are specified on their
        data. Indices for subspacing are then automatically inferred from
        where the conditions are met.

        Metadata constructs and the conditions on their data are defined
        by keyword parameters.

        * Any domain axes that have not been identified remain unchanged.

        * Multiple domain axes may be subspaced simultaneously, and it
          doesn't matter which order they are specified in.

        * Explicit indices may also be assigned to a domain axis
          identified by a metadata construct, with either a Python `slice`
          object, or a sequence of integers or booleans.

        * For a dimension that is cyclic, a subspace defined by a slice or
          by a `Query` instance is assumed to "wrap" around the edges of
          the data.

        * Conditions may also be applied to multi-dimensional metadata
          constructs. The "compress" mode is still the default mode (see
          the positional arguments), but because the indices may not be
          acting along orthogonal dimensions, some missing data may still
          need to be inserted into the field construct's data.

        .. versionadded:: 3.11.0

        .. seealso:: `indices`

        :Parameters:

            mode: `str`, *optional*
                There are two modes of operation, each of which provides
                indices for a different type of subspace:

                ==============  ==========================================
                *mode*          Description
                ==============  ==========================================
                ``'compress'``  Return indices that identify only the
                                requested locations.

                                This is the default mode.

                                Note that if a multi-dimensional metadata
                                construct is being used to define the
                                indices then some unrequested locations
                                may also be selected.

                ``'envelope'``  The returned subspace is the smallest that
                                contains all of the requested locations.

                ``'test'``      May be used on its own or in addition to
                                one of the other positional arguments. Do
                                not create a subspace, but return `True`
                                or `False` depending on whether or not it
                                is possible to create the specified
                                subspace.
                ==============  ==========================================

            kwargs: *optional*
                A keyword name is an identity of a metadata construct, and
                the keyword value provides a condition for inferring
                indices that apply to the dimension (or dimensions)
                spanned by the metadata construct's data. Indices are
                created that select every location for which the metadata
                construct's data satisfies the condition.

        :Returns:

            `Domain` or `bool`
                An independent domain construct containing the subspace of
                the original domain. If the ``'test'`` positional argument
                has been set then return `True` or `False` depending on
                whether or not it is possible to create specified
                subspace.

        **Examples**

        >>> d = cf.example_field(0).domain
        >>> print(d)
        Dimension coords: latitude(5) = [-75.0, ..., 75.0] degrees_north
                        : longitude(8) = [22.5, ..., 337.5] degrees_east
                        : time(1) = [2019-01-01 00:00:00]
        >>> print(d.subspace(X=112.5))
        Dimension coords: latitude(5) = [-75.0, ..., 75.0] degrees_north
                        : longitude(1) = [112.5] degrees_east
                        : time(1) = [2019-01-01 00:00:00]

        >>> print(d.indices(X=112.5, Y=cf.wi(-60, 30)))
        Dimension coords: latitude(2) = [-45.0, 0.0] degrees_north
                        : longitude(1) = [112.5] degrees_east
                        : time(1) = [2019-01-01 00:00:00]

        >>> print(d.indices(X=[-1, 0], Y=slice(1, -1))
        Dimension coords: latitude(3) = [-45.0, 0.0, 45.0] degrees_north
                       : longitude(2) = [337.5, 22.5] degrees_east
                       : time(1) = [2019-01-01 00:00:00]

        """
        test = False
        if "test" in mode:
            mode = list(mode)
            mode.remove("test")
            test = True

        if not mode and not kwargs:
            if test:
                return True

            return self.copy()

        try:
            indices = self.indices(*mode, **kwargs)
        except ValueError as error:
            if test:
                return False

            raise ValueError(error)

        if test:
            return True

        domain_axes = self.domain_axes(todict=True)

        axes = []
        shape = []
        indices2 = []
        for a, b in indices.items():
            axes.append(a)
            shape.append(domain_axes[a].get_size())
            indices2.append(b)

        indices, roll = parse_indices(
            tuple(shape), tuple(indices2), cyclic=True
        )

        if roll:
            new = self
            cyclic_axes = self.cyclic()
            for iaxis, shift in roll.items():
                axis = axes[iaxis]
                if axis not in cyclic_axes:
                    raise IndexError(
                        "Can't take a cyclic slice from non-cyclic "
                        f"{self.constructs.domain_axis_identity(axis)!r} "
                        "axis"
                    )

                new = new.roll(axis, shift)
        else:
            new = self.copy()

        # ------------------------------------------------------------
        # Set sizes of domain axes
        # ------------------------------------------------------------
        domain_axes = new.domain_axes(todict=True)
        for axis, size in zip(axes, indices_shape(indices, shape)):
            domain_axes[axis].set_size(size)

        # ------------------------------------------------------------
        # Subspace constructs that have data
        # ------------------------------------------------------------
        construct_data_axes = new.constructs.data_axes()

        for key, construct in new.constructs.filter_by_data().items():
            construct_axes = construct_data_axes[key]
            dice = [indices[axes.index(axis)] for axis in construct_axes]

            # Replace existing construct with its subspace
            new.set_construct(
                construct[tuple(dice)],
                key=key,
                axes=construct_axes,
                copy=False,
            )

        return new

    @_inplace_enabled(default=False)
    def transpose(self, axes, inplace=False):
        """Permute the data axes of the metadata constructs.

        Each metadata construct has its data axis order changed to the
        relative ordering defined by the *axes* parameter. For
        instance, if the given *axes* are ``['X', 'Z', 'Y']`` then a
        metadata construct whose data axis order is ``('Y', 'X')``
        will be transposed to have data order ``('X', 'Y')``.

        .. versionadded:: 3.11.0

        .. seealso:: `domain_axis`, `flip`

        :Parameters:

            axes: sequence of `str`
                Define the new domain axis order.

                A domain axis is identified by that which would be
                selected by passing a given axis description to a call of
                the `domain_axis` method. For example, a value of ``'X'``
                would select the domain axis construct returned by
                ``f.domain_axis('X')``.

                Each domain axis of the domain construct data must be
                specified.

            constructs: `bool`, optional
                If True then metadata constructs are also transposed so
                that their axes are in the same relative order as in the
                transposed data array of the field. By default metadata
                constructs are not altered.

            {{inplace: `bool`, optional}}

        :Returns:

            `Domain` or `None`
                The domain construct with transposed constructs, or `None`
                if the operation was in-place.

        **Examples**

        >>> d = cf.example_field(7).domain
        >>> print(d)
        Dimension coords: time(3) = [1979-05-01 12:00:00, 1979-05-02 12:00:00, 1979-05-03 12:00:00] gregorian
                        : air_pressure(1) = [850.0] hPa
                        : grid_latitude(4) = [0.44, ..., -0.88] degrees
                        : grid_longitude(5) = [-1.18, ..., 0.58] degrees
        Auxiliary coords: latitude(grid_latitude(4), grid_longitude(5)) = [[52.4243, ..., 51.1163]] degrees_north
                        : longitude(grid_latitude(4), grid_longitude(5)) = [[8.0648, ..., 10.9238]] degrees_east
        Coord references: grid_mapping_name:rotated_latitude_longitude


        >>> print(d.transpose(['X', 'T', 'Y', 'Z']))
        Dimension coords: time(3) = [1979-05-01 12:00:00, 1979-05-02 12:00:00, 1979-05-03 12:00:00] gregorian
                        : air_pressure(1) = [850.0] hPa
                        : grid_latitude(4) = [0.44, ..., -0.88] degrees
                        : grid_longitude(5) = [-1.18, ..., 0.58] degrees
        Auxiliary coords: latitude(grid_longitude(5), grid_latitude(4)) = [[52.4243, ..., 51.1163]] degrees_north
                        : longitude(grid_longitude(5), grid_latitude(4)) = [[8.0648, ..., 10.9238]] degrees_east
        Coord references: grid_mapping_name:rotated_latitude_longitude

        """
        d = _inplace_enabled_define_and_cleanup(self)

        # Parse the axes
        if axes is None:
            raise ValueError(
                f"Can't transpose {self.__class__.__name__}. "
                f"Must provide an order for all axes. Got: {axes}"
            )

        axes = d._parse_axes(axes)

        rank = self.rank
        if len(set(axes)) != rank:
            raise ValueError(
                f"Can't transpose {self.__class__.__name__}. "
                f"Must provide an unambiguous order for all "
                f"{rank} domain axes. Got: {axes}"
            )

        data_axes = d.constructs.data_axes()
        for key, construct in d.constructs.filter_by_data().items():
            construct_axes = data_axes[key]

            if len(construct_axes) < 2:
                # No need to transpose 1-d constructs
                continue

            # Transpose the construct
            iaxes = [
                construct_axes.index(a) for a in axes if a in construct_axes
            ]
            construct.transpose(iaxes, inplace=True)

            # Update the axis order
            new_axes = [construct_axes[i] for i in iaxes]
            d.set_data_axes(axes=new_axes, key=key)

        return d



import atexit
import datetime
import faulthandler
import os
import tempfile
import unittest

import numpy as np

faulthandler.enable()  # to debug seg faults and timeouts

import cf

n_tmpfiles = 1
tmpfiles = [
    tempfile.mkstemp("_test_pp.nc", dir=os.getcwd())[1]
    for i in range(n_tmpfiles)
]
[tmpfile] = tmpfiles


def _remove_tmpfiles():
    """Try to remove defined temporary files by deleting their paths."""
    for f in tmpfiles:
        try:
            os.remove(f)
        except OSError:
            pass


atexit.register(_remove_tmpfiles)


class ppTest(unittest.TestCase):
    ppfile = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "wgdos_packed.pp"
    )

    ppextradata = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "extra_data.pp"
    )

    new_table = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "new_STASH_to_CF.txt"
    )
    text_file = open(new_table, "w")
    text_file.write(
        "1!30201!long name                           !Pa!!!NEW_NAME!!"
    )
    text_file.close()

    chunk_sizes = (800000, 80000)

    def test_PP_read_um(self):
        f = cf.read(self.ppextradata)[0]

        g = cf.read(self.ppextradata, um={"fmt": "pp"})[0]
        self.assertTrue(f.equals(g))

        for vn in (4.5, 405, "4.5", None):
            g = cf.read(self.ppextradata, um={"fmt": "pp", "version": vn})[0]
            self.assertTrue(f.equals(g))

        p = cf.read("wgdos_packed.pp")[0]
        p0 = cf.read(
            "wgdos_packed.pp",
            um={
                "fmt": "PP",
                "endian": "little",
                "word_size": 4,
                "version": 4.5,
                "height_at_top_of_model": 23423.65,
            },
        )[0]

        self.assertTrue(p.equals(p0, verbose=2))

    def test_load_stash2standard_name(self):
        f = cf.read(self.ppfile)[0]
        self.assertEqual(f.identity(), "eastward_wind")
        self.assertEqual(f.Units, cf.Units("m s-1"))

        for merge in (True, False):
            cf.load_stash2standard_name(self.new_table, merge=merge)
            f = cf.read(self.ppfile)[0]
            self.assertEqual(f.identity(), "NEW_NAME")
            self.assertEqual(f.Units, cf.Units("Pa"))
            cf.load_stash2standard_name()
            f = cf.read(self.ppfile)[0]
            self.assertEqual(f.identity(), "eastward_wind")
            self.assertEqual(f.Units, cf.Units("m s-1"))

        cf.load_stash2standard_name()

    def test_stash2standard_name(self):
        d = cf.stash2standard_name()
        self.assertIsInstance(d, dict)
        d["test"] = None
        e = cf.stash2standard_name()
        self.assertNotEqual(d, e)

    def test_PP_read_select(self):
        f = cf.read(self.ppfile, select="lbproc=0")
        self.assertEqual(len(f), 1)

    def test_PP_WGDOS_UNPACKING(self):
        f = cf.read(self.ppfile)[0]

        self.assertEqual(f.data.mean(), 3.8080420658506196)

        array = f.array

        f = cf.read(self.ppfile)[0]

        # TODO: reinstate "CFA4" at version>3.14
        for fmt in ("NETCDF4",):  # "CFA4"):
            cf.write(f, tmpfile, fmt=fmt)
            g = cf.read(tmpfile)[0]

            self.assertTrue((f.array == array).all())
            self.assertTrue(f.equals(g, verbose=2))

    def test_PP_extra_data(self):
        f = cf.read(self.ppextradata)[0]

        self.assertEqual(len(f.dimension_coordinates()), 3)
        self.assertEqual(len(f.auxiliary_coordinates()), 3)

        sites = f.dimension_coordinate("long_name=site")
        self.assertTrue(np.allclose(sites, [1, 2, 3]))

        regions = f.auxiliary_coordinate("region").array
        self.assertEqual(
            regions.tolist(),
            ["Northern Hemisphere", "Southern Hemisphere", "Global"],
        )

        self.assertTrue(f.dimension_coordinate("height", default=False))
        self.assertTrue(f.dimension_coordinate("time", default=False))
        self.assertTrue(f.auxiliary_coordinate("longitude", default=False))


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import cfdm

from . import Bounds, mixin


class DomainAncillary(mixin.PropertiesDataBounds, cfdm.DomainAncillary):
    """A domain ancillary construct of the CF data model.

    A domain ancillary construct provides information which is needed
    for computing the location of cells in an alternative coordinate
    system. It is referenced by a term of a coordinate conversion
    formula of a coordinate reference construct. It contains a data
    array which depends on zero or more of the domain axes.

    It also contains an optional array of cell bounds, stored in a
    `cf.Bounds` object, recording the extents of each cell (only
    applicable if the array contains coordinate data), and properties
    to describe the data.

    An array of cell bounds spans the same domain axes as the data
    array, with the addition of an extra dimension whose size is that
    of the number of vertices of each cell.

    **NetCDF interface**

    {{netcdf variable}}

    """

    def __new__(cls, *args, **kwargs):
        """Store component classes."""
        instance = super().__new__(cls)
        instance._Bounds = Bounds
        return instance

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        return super().__repr__().replace("<", "<CF ", 1)



"""Functions used during the creation of `Data` objects."""
from functools import lru_cache

import dask.array as da
import numpy as np
from dask.base import is_dask_collection


def to_dask(array, chunks, **from_array_options):
    """Create a `dask` array.

    .. versionadded:: TODODASKVER

    :Parameters:

        array: array_like
            The array to be converted to a `dask` array. Examples of
            valid types include `numpy` arrays, `dask` arrays, `Array`
            subclasses, `list`, `tuple`, scalars.

        chunks: `int`, `tuple`, `dict` or `str`, optional
            Specify the chunking of the returned dask array.  Any
            value accepted by the *chunks* parameter of the
            `dask.array.from_array` function is allowed.

            Ignored if *array* is a `dask` array, which already
            defines its own chunks.

            Might get automatically modified if *array* is a
            compressed `Array` subclass.

        from_array_options: `dict`, optional
            Keyword arguments to be passed to `dask.array.from_array`.

            If *from_array_options* has no ``'lock'`` key then the
            `lock` keyword is set to the `_dask_lock` attribute of
            *array* or, if there is no such attribute, `False`.

            If *from_array_options* has no ``'meta'`` key then the
            `meta` keyword is set to the `_dask_meta` attribute of
            *array* or, if there is no such attribute, `None`.

    :Returns:

        `dask.array.Array`
            The `dask` array representation of the array.

    **Examples**

    >>> cf.data.creation.to_dask([1, 2, 3], 'auto')
    dask.array<array, shape=(3,), dtype=int64, chunksize=(3,), chunktype=numpy.ndarray>
    >>> cf.data.creation.to_dask([1, 2, 3], chunks=2)
    dask.array<array, shape=(3,), dtype=int64, chunksize=(2,), chunktype=numpy.ndarray>
    >>> cf.data.creation.to_dask([1, 2, 3], chunks=2, {'asarray': True})
    dask.array<array, shape=(3,), dtype=int64, chunksize=(2,), chunktype=numpy.ndarray>
    >>> cf.data.creation.to_dask(cf.dt(2000, 1, 1), 'auto')
    dask.array<array, shape=(), dtype=object, chunksize=(), chunktype=numpy.ndarray>
    >>> cf.data.creation.to_dask([cf.dt(2000, 1, 1)], 'auto')
    dask.array<array, shape=(1,), dtype=object, chunksize=(1,), chunktype=numpy.ndarray>

    """
    if is_dask_collection(array):
        return array

    if hasattr(array, "to_dask_array"):
        try:
            return array.to_dask_array(chunks=chunks)
        except TypeError:
            return array.to_dask_array()

    if not isinstance(
        array, (np.ndarray, list, tuple, memoryview) + np.ScalarType
    ) and not hasattr(array, "shape"):
        # 'array' is not of a type that `da.from_array` can cope with,
        # so convert it to a numpy array.
        array = np.asanyarray(array)

    kwargs = from_array_options
    kwargs.setdefault("lock", getattr(array, "_dask_lock", False))
    kwargs.setdefault("meta", getattr(array, "_dask_meta", None))

    try:
        return da.from_array(array, chunks=chunks, **kwargs)
    except NotImplementedError:
        # Try again with 'chunks=-1', in case the failure was due to
        # not being able to use auto rechunking with object dtype.
        return da.from_array(array, chunks=-1, **kwargs)


@lru_cache(maxsize=32)
def generate_axis_identifiers(n):
    """Return new axis identifiers for a given number of axes.

    The names are arbitrary and have no semantic meaning.

    .. versionadded:: TODODASKVER

    :Parameters:

        n: `int`
            Generate this number of axis identifiers.

    :Returns:

        `list`
            The new axis identifiers.

    **Examples**

    >>> cf.data.creation.generate_axis_identifiers(0)
    []
    >>> cf.data.creation.generate_axis_identifiers(1)
    ['dim0']
    >>> cf.data.creation.generate_axis_identifiers(3)
    ['dim0', 'dim1', 'dim2']

    """
    return [f"dim{i}" for i in range(n)]



import cfdm
import numpy as np

from . import Bounds, mixin
from .data.data import Data
from .decorators import (
    _deprecated_kwarg_check,
    _inplace_enabled,
    _inplace_enabled_define_and_cleanup,
)
from .functions import _DEPRECATION_ERROR_ATTRIBUTE, _DEPRECATION_ERROR_KWARGS
from .timeduration import TimeDuration
from .units import Units


class DimensionCoordinate(
    mixin.Coordinate, mixin.PropertiesDataBounds, cfdm.DimensionCoordinate
):
    """A dimension coordinate construct of the CF data model.

    A dimension coordinate construct provides information which locate
    the cells of the domain and which depend on a subset of the domain
    axis constructs. The dimension coordinate construct is able to
    unambiguously describe cell locations because a domain axis can be
    associated with at most one dimension coordinate construct, whose
    data array values must all be non-missing and strictly
    monotonically increasing or decreasing. They must also all be of
    the same numeric data type. If cell bounds are provided, then each
    cell must have exactly two vertices. CF-netCDF coordinate
    variables and numeric scalar coordinate variables correspond to
    dimension coordinate constructs.

    The dimension coordinate construct consists of a data array of the
    coordinate values which spans a subset of the domain axis
    constructs, an optional array of cell bounds recording the extents
    of each cell (stored in a `Bounds` object), and properties to
    describe the coordinates. An array of cell bounds spans the same
    domain axes as its coordinate array, with the addition of an extra
    dimension whose size is that of the number of vertices of each
    cell. This extra dimension does not correspond to a domain axis
    construct since it does not relate to an independent axis of the
    domain. Note that, for climatological time axes, the bounds are
    interpreted in a special way indicated by the cell method
    constructs.

    **NetCDF interface**

    {{netcdf variable}}

    """

    def __new__(cls, *args, **kwargs):
        """Store component classes."""
        instance = super().__new__(cls)
        instance._Bounds = Bounds
        return instance

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        return super().__repr__().replace("<", "<CF ", 1)

    def _centre(self, period):
        """It assumed, but not checked, that the period has been set.

        .. seealso:: `roll`

        """
        if self.direction():
            mx = self.data[-1]
        else:
            mx = self.data[0]

        return ((mx // period) * period).squeeze()

    def _infer_direction(self):
        """Return True if a coordinate is increasing, otherwise return
        False.

        A dimension coordinate construct is considered to be increasing if
        its data array values are increasing in index space, or if it has
        no data nor bounds.

        If the direction can not be inferred from the data not bounds then
        the coordinate's units are used.

        The direction is inferred from the coordinate's data array values
        or its from coordinates. It is not taken directly from its
        `cf.Data` object.

        :Returns:

            `bool`
                Whether or not the coordinate is increasing.

        **Examples**

        >>> c.array
        array([  0  30  60])
        >>> c._get_direction()
        True
        >>> c.array
        array([15])
        >>> c.bounds.array
        array([  30  0])
        >>> c._get_direction()
        False

        """
        data = self.get_data(None, _fill_value=False)
        if data is not None:
            # Infer the direction from the data
            if data.size > 1:
                data = data[0:2].array
                return bool(data.item(0) < data.item(1))

        # Still here?
        data = self.get_bounds_data(None, _fill_value=False)
        if data is not None:
            # Infer the direction from the bounds
            b = data[(0,) * (data.ndim - 1)].array
            return bool(b.item(0) < b.item(1))

        # Still here? Then infer the direction from the units.
        return not self.Units.ispressure

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    #    @property
    #    def cellsize(self):
    #        '''The cell sizes.
    #
    #    :Returns:
    #
    #        `Data`
    #            The size for each cell.
    #
    #    **Examples**
    #
    #    >>> print(c.bounds)
    #    <CF Bounds: latitude(3, 2) degrees_north>
    #    >>> print(c.bounds.array)
    #    [[-90. -87.]
    #     [-87. -80.]
    #     [-80. -67.]]
    #    >>> print(d.cellsize)
    #    <CF Data(3): [3.0, 7.0, 13.0] degrees_north>
    #    >>> print(d.cellsize.array)
    #    [  3.   7.  13.]
    #    >>> print(c.sin().cellsize.array)
    #    [ 0.00137047  0.01382178  0.0643029 ]
    #
    #    >>> del(c.bounds)
    #    >>> c.cellsize
    #    AttributeError: Can't get cell sizes when coordinates have no bounds
    #
    #        '''
    #        cells = self.get_bounds_data(None)
    #        if cells is None:
    #            raise AttributeError(
    #                "Can't get cell sizes when coordinates have no bounds")
    #
    #        if self.direction():
    #            cells = cells[:, 1] - cells[:, 0]
    #        else:
    #            cells = cells[:, 0] - cells[:, 1]
    #
    #        cells.squeeze(1, inplace=True)
    #
    #        return cells

    @property
    def decreasing(self):
        """True if the dimension coordinate is decreasing, otherwise
        False.

        A dimension coordinate is increasing if its coordinate values are
        increasing in index space.

        The direction is inferred from one of, in order of precedence:

        * The data array
        * The bounds data array
        * The `units` CF property

        :Returns:

            `bool`
                Whether or not the coordinate is decreasing.

        **Examples**

        >>> c.decreasing
        False
        >>> c.flip().increasing
        True

        """
        return not self.direction()

    @property
    def increasing(self):
        """`True` for dimension coordinate constructs, `False`
        otherwise.

        A dimension coordinate is increasing if its coordinate values are
        increasing in index space.

        The direction is inferred from one of, in order of precedence:

        * The data array
        * The bounds data array
        * The `units` CF property

        :Returns:

            `bool`
                Whether or not the coordinate is increasing.

        **Examples**

        >>> c.decreasing
        False
        >>> c.flip().increasing
        True

        """
        return self.direction()

    #    @property
    #    def lower_bounds(self):
    #        '''The lower dimension coordinate bounds in a `cf.Data` object.
    #
    #    .. seealso:: `bounds`, `upper_bounds`
    #
    #    **Examples**
    #
    #    >>> print(c.bounds.array)
    #    [[ 5  3]
    #     [ 3  1]
    #     [ 1 -1]]
    #    >>> c.lower_bounds
    #    <CF Data(3): [3, 1, -1]>
    #    >>> print(c.lower_bounds.array)
    #    [ 3  1 -1]
    #
    #        '''
    #        data = self.get_bounds_data(None)
    #        if data is None:
    #            raise ValueError(
    #                "Can't get lower bounds when there are no bounds")
    #
    #        if self.direction():
    #            i = 0
    #        else:
    #            i = 1
    #
    #        out = data[..., i]
    #        out.squeeze(1, inplace=True)
    #        return out

    #
    #    @property
    #    def upper_bounds(self):
    #        '''The upper dimension coordinate bounds in a `cf.Data` object.
    #
    #    .. seealso:: `bounds`, `lower_bounds`
    #
    #    **Examples**
    #
    #    >>> print(c.bounds.array)
    #    [[ 5  3]
    #     [ 3  1]
    #     [ 1 -1]]
    #    >>> c.upper_bounds
    #    <CF Data(3): [5, 3, 1]>
    #    >>> print(c.upper_bounds.array)
    #    [5  3  1]
    #
    #        '''
    #        data = self.get_bounds_data(None)
    #        if data is None:
    #            raise ValueError(
    #                "Can't get upper bounds when there are no bounds")
    #
    #        if self.direction():
    #            i = 1
    #        else:
    #            i = 0
    #
    #        return data[..., i].squeeze(1)

    def direction(self):
        """Return True if the dimension coordinate values are
        increasing, otherwise return False.

        Dimension coordinates values are increasing if its coordinate
        values are increasing in index space.

        The direction is inferred from one of, in order of precedence:

        * The data array
        * The bounds data array
        * The `units` CF property

        :Returns:

            `bool`
                Whether or not the coordinate is increasing.

        **Examples**

        >>> c.array
        array([  0  30  60])
        >>> c.direction()
        True

        >>> c.bounds.array
        array([  30  0])
        >>> c.direction()
        False

        """
        _direction = self._custom.get("direction")
        if _direction is not None:
            return _direction

        _direction = self._infer_direction()
        self._custom["direction"] = _direction

        return _direction

    def create_bounds(
        self, bound=None, cellsize=None, flt=0.5, max=None, min=None
    ):
        """Create cell bounds.

        Creates new cell bounds, irrespective of whether the cells
        already have cell bounds. The new bounds are not set on the
        dimension coordinate construct, but if that is desired they
        may always be added with the `set_bounds` method, for
        instance:

        >>> b = d.create_bounds()
        >>> d.set_bounds(b)

        By default, Voronoi cells are created by defining cell bounds
        that are half way between adjacent coordinate values. For
        non-cyclic coordinates, extrapolation is used so that the
        first and last cells have their coordinates in their centres.

        For instance, the Voronoi cells created for non-cyclic
        coordinates ``0, 40, 90, 180`` have bounds ``(-20, 20), (20,
        65), (65, 135), (135, 225)``. If those coordinates were
        instead cyclic with a period of ``360``, then the bounds would
        be ``(-90, 20), (20, 65), (65, 135), (135, 270)``.

        .. seealso:: `del_bounds`, `get_bounds`, `set_bounds`

        :Parameters:

            bound: data_like scalar, optional
                If set to a value larger (smaller) than the largest
                (smallest) coordinate value, then bounds are created
                which include this value and for which each coordinate
                is in the centre of its bounds. This can be useful for
                recreating the bounds of vertical coordinates from
                Earth system models.

                *Parameter example:*
                  For coordinates ``1, 2, 10``, setting ``bound=0.5``
                  will result in bounds of ``(0.5, 1.5), (1.5, 2.5),
                  (2.5, 17.5)``.

                *Parameter example:*
                  Beware that if *bound* is too large or small, then
                  some returned bounds values will by physically
                  incorrect. For coordinates ``1, 2, 10``, the setting
                  of ``bound=-2`` will result in bounds of ``(-2, 4),
                  (4, 0), (0, 20)``.

            cellsize: optional
                Define the size of each cell that is created. Created
                cells are allowed to overlap and do not have to be
                contigious. The *cellsize* parameter may be one of:

                * data_like scalar

                  Defines the cell size, either in the same units as
                  the coordinates or in the units provided. Note that
                  in this case, the position of each coordinate within
                  the its cell is determined by the *flt* parameter.

            ..

                * `cf.TimeDuration`

                  A time duration defining the cell size. Only
                  applicable to reference time coordinates. It is
                  possible to "anchor" the cell bounds via the
                  `cf.TimeDuration` instance parameters. For example,
                  to specify cell size of one calendar month, starting
                  and ending on the 15th day:
                  ``cellsize=cf.M(day=15)`` (see `cf.M` for
                  details). Note that the *flt* parameter is ignored
                  in this case.

            flt: array_like scalar, optional
                When creating cells with sizes specified by the
                *cellsize* parameter, define the fraction of each
                cell which is less than its coordinate value. By default
                *flt* is 0.5, so that each cell has its coordinate at
                its centre. Ignored if *cellsize* is not set.

                *Parameter example:*
                  For coordinates ``1, 2, 10`` setting ``flt=0.25``
                  will result in bounds of ``(0.75, 1.75), (1.75,
                  2.75), (9.75, 10.75)``.

                *Parameter example:*
                  For decreasing coordinates ``2, 0, -12``, setting
                  ``cellsize=6, flt=0.9`` will result in bounds of
                  ``(2.6, -3.4), (0.6, -5.4), (-11.4, -17.4)``.

            max: data_like scalar, optional
                Limit the created bounds to be no more than this number.

                *Parameter example:*
                  To ensure that all latitude bounds are at most
                  ``90``: ``max=90``, or ``max=cf.Data(90,
                  'degrees_north')``.

            min: data_like scalar, optional
                Limit the created bounds to be no less than this number.

                *Parameter example:*
                  To ensure that all latitude bounds are at least
                  ``-90``: ``min=-90``, or ``min=cf.Data(-90,
                  'degrees_north')``.

        :Returns:

            `Bounds`
                The new coordinate cell bounds.

        **Examples**

        Create bounds for Voronoi cells:

        >>> d = cf.DimensionCoordinate(data=[0.0, 40, 90, 180])
        >>> b = d.create_bounds()
        >>> print(b.array)
        [[-20.  20.]
         [ 20.  65.]
         [ 65. 135.]
         [135. 225.]]
        >>> d = d[::-1]
        >>> b = d.create_bounds()
        >>> print(b.array)
        [[225. 135.]
         [135.  65.]
         [ 65.  20.]
         [ 20. -20.]]

        Cyclic coordinates have a co-located first and last bound:

        >>> d.period(360)
        >>> d.cyclic(0)
        set()
        >>> b = d.create_bounds()
        >>> print(b.array)
        [[270. 135.]
         [135.  65.]
         [ 65.  20.]
         [ 20. -90.]]

        Cellsize units must be equivalent to the coordinate units, or
        if the cell has no units then they are assumed to be the
        same as the coordinates:

        >>> d = cf.DimensionCoordinate(data=cf.Data([0, 2, 4], 'km'))
        >>> b = d.create_bounds(cellsize=cf.Data(2000, 'm'))
        >>> print(b.array)
        [[-1  1]
         [ 1  3]
         [ 3  5]]
        >>> b = d.create_bounds(cellsize=2)
        >>> print(b.array)
        [[-1  1]
         [ 1  3]
         [ 3  5]]

        Cells may be non-contiguous and may overlap:

        >>> d = cf.DimensionCoordinate(data=[1.0, 2, 10])
        >>> b = d.create_bounds(cellsize=1)
        >>> print(b.array)
        [[ 0.5  1.5]
         [ 1.5  2.5]
         [ 9.5 10.5]]
        >>> b = d.create_bounds(cellsize=5)
        >>> print(b.array)
        [[-1.5  3.5]
         [-0.5  4.5]
         [ 7.5 12.5]]

        Reference date-time coordinates can use `cf.TimeDuration` cell
        sizes:

        >>> d = cf.DimensionCoordinate(
        ...   data=cf.Data(['1984-12-01', '1984-12-02'], dt=True)
        ... )
        >>> b = d.create_bounds(cellsize=cf.D())
        >>> print(b.datetime_array)
        [[cftime.DatetimeGregorian(1984, 12, 1, 0, 0, 0, 0)
          cftime.DatetimeGregorian(1984, 12, 2, 0, 0, 0, 0)]
         [cftime.DatetimeGregorian(1984, 12, 2, 0, 0, 0, 0)
          cftime.DatetimeGregorian(1984, 12, 3, 0, 0, 0, 0)]]
        >>> b = d.create_bounds(cellsize=cf.D(hour=12))
        >>> print(b.datetime_array)
        [[cftime.DatetimeGregorian(1984, 11, 30, 12, 0, 0, 0)
          cftime.DatetimeGregorian(1984, 12, 1, 12, 0, 0, 0)]
         [cftime.DatetimeGregorian(1984, 12, 1, 12, 0, 0, 0)
          cftime.DatetimeGregorian(1984, 12, 2, 12, 0, 0, 0)]]

        Cell bounds defined by calendar months:

        >>> d = cf.DimensionCoordinate(
        ...   data=cf.Data(['1985-01-16: 12:00', '1985-02-15'], dt=True)
        ... )
        >>> b = d.create_bounds(cellsize=cf.M())
        >>> print(b.datetime_array)
        [[cftime.DatetimeGregorian(1985, 1, 1, 0, 0, 0, 0)
          cftime.DatetimeGregorian(1985, 2, 1, 0, 0, 0, 0)]
         [cftime.DatetimeGregorian(1985, 2, 1, 0, 0, 0, 0)
          cftime.DatetimeGregorian(1985, 3, 1, 0, 0, 0, 0)]]
        >>> b = d.create_bounds(cellsize=cf.M(day=20))
        >>> print(b.datetime_array)
        [[cftime.DatetimeGregorian(1984, 12, 20, 0, 0, 0, 0
          cftime.DatetimeGregorian(1985, 1, 20, 0, 0, 0, 0]
         [cftime.DatetimeGregorian(1985, 1, 20, 0, 0, 0, 0
          cftime.DatetimeGregorian(1985, 2, 20, 0, 0, 0, 0]]

        Cell bounds defined by calendar years:

        >>> d = cf.DimensionCoordinate(
        ...   data=cf.Data(['1984-06-01', '1985-06-01'], dt=True)
        ... )
        >>> b = d.create_bounds(cellsize=cf.Y(month=12))
        >>> print(b.datetime_array)
        [[cftime.DatetimeGregorian(1983, 12, 1, 0, 0, 0, 0)
          cftime.DatetimeGregorian(1984, 12, 1, 0, 0, 0, 0)]
         [cftime.DatetimeGregorian(1984, 12, 1, 0, 0, 0, 0)
          cftime.DatetimeGregorian(1985, 12, 1, 0, 0, 0, 0)]]

        """
        array = self.array
        size = array.size

        if cellsize is not None:
            if bound:
                raise ValueError(
                    "bound parameter can't be True when setting the "
                    "cellsize parameter"
                )

            if not isinstance(cellsize, TimeDuration):
                # ----------------------------------------------------
                # Create bounds based on cell sizes defined by a
                # data-like object
                #
                # E.g. cellsize=10
                #      cellsize=cf.Data(1, 'day')
                # ----------------------------------------------------
                cellsize = Data.asdata(abs(cellsize))
                if cellsize.Units:
                    err_msg = (
                        "Can't create bounds because the bound units "
                        f"({cellsize.Units}) are not compatible with "
                        f"the coordinate units ({self.Units})."
                    )
                    if self.Units.isreftime:
                        if not cellsize.Units.istime:
                            raise ValueError(err_msg)
                        cellsize.Units = Units(self.Units._utime.units)
                    else:
                        if not cellsize.Units.equivalent(self.Units):
                            raise ValueError(err_msg)
                        cellsize.Units = self.Units
                cellsize = cellsize.datum()

                cellsize0 = cellsize * flt
                cellsize1 = cellsize * (1 - flt)
                if not self.direction():
                    cellsize0, cellsize1 = -cellsize1, -cellsize0

                bounds = np.empty((size, 2), dtype=array.dtype)
                bounds[:, 0] = array - cellsize0
                bounds[:, 1] = array + cellsize1
            else:
                # ----------------------------------------------------
                # Create bounds based on cell sizes defined by a
                # TimeDuration object
                #
                # E.g. cellsize=cf.s()
                #      cellsize=cf.m()
                #      cellsize=cf.h()
                #      cellsize=cf.D()
                #      cellsize=cf.M()
                #      cellsize=cf.Y()
                #      cellsize=cf.D(hour=12)
                #      cellsize=cf.M(day=16)
                #      cellsize=cf.M(2)
                #      cellsize=cf.M(2, day=15, hour=12)
                # ----------------------------------------------------
                if not self.Units.isreftime:
                    raise ValueError(
                        "Can't create reference time bounds for "
                        f"non-reference time coordinates: {self.Units!r}"
                    )

                bounds = np.empty((size, 2), dtype=object)

                cellsize_bounds = cellsize.bounds
                direction = bool(self.direction())

                for c, b in zip(self.datetime_array, bounds):
                    b[...] = cellsize_bounds(c, direction=direction)
        else:
            if bound is None:
                # ----------------------------------------------------
                # Creat Voronoi bounds
                # ----------------------------------------------------
                if size < 2:
                    raise ValueError(
                        "Can't create Voronoi bounds when there is "
                        "only one cell"
                    )

                bounds_1d = [array.item(0) * 1.5 - array.item(1) * 0.5]
                bounds_1d.extend((array[0:-1] + array[1:]) * 0.5)
                bounds_1d.append(array.item(-1) * 1.5 - array.item(-2) * 0.5)

                dtype = type(bounds_1d[0])

                if self.cyclic():
                    period = self.period()
                    if period is None:
                        raise ValueError(
                            "Can't create Voronoi bounds for cyclic "
                            "coordinates with no period"
                        )

                    if self.direction():
                        # Increasing
                        b = 0.5 * (array[0] + period - array[-1])
                    else:
                        # Decreasing
                        b = 0.5 * (array[-1] - period + array[0])

                    bounds_1d[0] = array.item(0) - b
                    bounds_1d[-1] = array.item(-1) + b
                else:
                    if max is not None:
                        if self.direction():
                            bounds_1d[-1] = max
                        else:
                            bounds_1d[0] = max
                    if min is not None:
                        if self.direction():
                            bounds_1d[0] = min
                        else:
                            bounds_1d[-1] = min
            else:
                # ----------------------------------------------------
                # Create
                # ----------------------------------------------------
                direction = self.direction()
                if not direction and size > 1:
                    array = array[::-1]

                bounds_1d = [bound]
                if bound <= array.item(0):
                    for i in range(size):
                        bound = 2.0 * array.item(i) - bound
                        bounds_1d.append(bound)
                elif bound >= array.item(-1):
                    for i in range(size - 1, -1, -1):
                        bound = 2.0 * array.item(i) - bound
                        bounds_1d.append(bound)

                    bounds_1d = bounds_1d[::-1]
                else:
                    raise ValueError("bad bound value")

                dtype = type(bounds_1d[-1])

                if not direction:
                    bounds_1d = bounds_1d[::-1]

            bounds = np.empty((size, 2), dtype=dtype)
            bounds[:, 0] = bounds_1d[:-1]
            bounds[:, 1] = bounds_1d[1:]

        # Create coordinate bounds object
        bounds = Bounds(data=Data(bounds, units=self.Units), copy=False)

        return bounds

    @_deprecated_kwarg_check("i", version="3.0.0", removed_at="4.0.0")
    @_inplace_enabled(default=False)
    def flip(self, axes=None, inplace=False, i=False):
        """Flips the dimension coordinate, that is reverses its
        direction."""
        d = _inplace_enabled_define_and_cleanup(self)

        super(DimensionCoordinate, d).flip(axes=axes, inplace=True)

        direction = d._custom.get("direction")
        if direction is not None:
            d._custom["direction"] = not direction

        return d

    def get_bounds(self, default=ValueError(), **kwargs):
        """Return the bounds.

        .. versionadded:: 3.0.0

        .. seealso:: `bounds`, `create_bounds', `get_data`,
                     `del_bounds`, `has_bounds`, `set_bounds`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if bounds
                have not been set.

                {{default Exception}}

        :Returns:

            `Bounds`
                The bounds.

        **Examples**

        >>> b = Bounds(data=cfdm.Data(range(10).reshape(5, 2)))
        >>> c.set_bounds(b)
        >>> c.has_bounds()
        True
        >>> c.get_bounds()
        <Bounds: (5, 2) >
        >>> b = c.del_bounds()
        >>> b
        <Bounds: (5, 2) >
        >>> c.has_bounds()
        False
        >>> print(c.get_bounds(None))
        None
        >>> print(c.del_bounds(None))
        None

        """
        if kwargs:
            _DEPRECATION_ERROR_KWARGS(
                self,
                "get_bounds",
                kwargs,
                "Bounds creation now uses the 'create_bounds' and "
                "'set_bounds' methods.",
            )  # pragma: no cover

        return super().get_bounds(default=default)

    #    def autoperiod(self, verbose=False):
    #        '''TODO Set dimensions to be cyclic.
    #
    #    TODO A dimension is set to be cyclic if it has a unique longitude (or
    #    grid longitude) dimension coordinate construct with bounds and the
    #    first and last bounds values differ by 360 degrees (or an
    #    equivalent amount in other units).
    #
    #    .. versionadded:: 3.0.0
    #
    #    .. seealso:: `isperiodic`, `period`
    #
    #    :Parameters:
    #
    #        TODO
    #
    #    :Returns:
    #
    #       `bool`
    #
    #    **Examples**
    #
    #    >>> f.autocyclic()
    #
    #        '''
    #        if not self.Units.islongitude:
    #            if verbose:
    #                print(0)
    #            if (self.get_property('standard_name', None) not in
    #                    ('longitude', 'grid_longitude')):
    #                if verbose:
    #                    print(1)
    #                return False
    #
    #        bounds = self.get_bounds(None)
    #        if bounds is None:
    #            if verbose:
    #                print(2)
    #            return False
    #
    #        bounds_data = bounds.get_data(None)
    #        if bounds_data is None:
    #            if verbose:
    #                print(3)
    #            return False
    #
    #        bounds = bounds_data.array
    #
    #        period = Data(360.0, units='degrees')
    #
    #        period.Units = bounds_data.Units
    #
    #        if abs(bounds[-1, -1] - bounds[0, 0]) != period.array:
    #            if verbose:
    #                print(4)
    #            return False
    #
    #        self.period(period)
    #
    #        return True

    @_deprecated_kwarg_check("i", version="3.0.0", removed_at="4.0.0")
    @_inplace_enabled(default=False)
    def roll(self, axis, shift, inplace=False, i=False):
        """Rolls the dimension coordinate along a cyclic axis."""
        c = _inplace_enabled_define_and_cleanup(self)

        size = c.size
        if size <= 1:
            return c

        shift %= size
        if not shift:
            # Null roll
            return c

        #        period = self._custom.get('period')
        period = c.period()

        if period is None:
            raise ValueError(
                f"Can't roll {self.__class__.__name__} when no period has "
                "been set"
            )

        direction = c.direction()

        centre = c._centre(period)

        if axis not in [0, -1]:
            raise ValueError(
                f"Can't roll axis {axis} when there is only one axis"
            )

        super(DimensionCoordinate, c).roll(axis, shift, inplace=True)

        c.dtype = np.result_type(c.dtype, period.dtype)

        data = c.data

        b = c.get_bounds(None)
        bounds_data = c.get_bounds_data(None, _fill_value=False)

        if bounds_data is not None:
            b.dtype = np.result_type(bounds_data.dtype, period.dtype)
            bounds_data = b.get_data(None, _fill_value=False)

        if direction:
            # Increasing
            data[:shift] -= period
            if bounds_data is not None:
                bounds_data[:shift] -= period

            if data[0] <= centre - period:
                data += period
                if bounds_data is not None:
                    bounds_data += period
        else:
            # Decreasing
            data[:shift] += period
            if bounds_data is not None:
                bounds_data[:shift] += period

            if data[0] >= centre + period:
                data -= period
                if bounds_data is not None:
                    bounds_data -= period

        c._custom["direction"] = direction

        return c

    # ----------------------------------------------------------------
    # Deprecated attributes and methods
    # ----------------------------------------------------------------
    @property
    def role(self):
        """Deprecated at version 3.0.0, use `construct_type` attribute
        instead."""
        _DEPRECATION_ERROR_ATTRIBUTE(
            self,
            "role",
            "Use attribute 'construct_type' instead",
            version="3.0.0",
            removed_at="4.0.0",
        )  # pragma: no cover



import numpy as np

from .abstract import Array


class FullArray(Array):
    """A array filled with a given value.

    The array may be empty or all missing values.

    """

    def __init__(
        self,
        fill_value=None,
        dtype=None,
        shape=None,
        units=False,
        calendar=False,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            fill_value : scalar, optional
                The fill value for the array. May be set to
                `cf.masked` or `np.ma.masked`.

            dtype: `numpy.dtype`
                The data type of the array.

            shape: `tuple`
                The array dimension sizes.

            units: `str` or `None`, optional
                The units of the netCDF variable. Set to `None` to
                indicate that there are no units. If unset then the
                units will be set to `None` during the first
                `__getitem__` call.

            calendar: `str` or `None`, optional
                The calendar of the netCDF variable. By default, or if
                set to `None`, then the CF default calendar is
                assumed, if applicable. If unset then the calendar
                will be set to `None` during the first `__getitem__`
                call.

            source: optional
                Initialise the array from the given object.

                {{init source}}

            {{deep copy}}

        """
        super().__init__(source=source, copy=copy)

        if source is not None:
            try:
                fill_value = source._get_component("fill_value", None)
            except AttributeError:
                fill_value = None

            try:
                dtype = source._get_component("dtype", None)
            except AttributeError:
                dtype = None

            try:
                shape = source._get_component("shape", None)
            except AttributeError:
                shape = None

            try:
                units = source._get_component("units", False)
            except AttributeError:
                units = False

            try:
                calendar = source._get_component("calendar", None)
            except AttributeError:
                calendar = None

        self._set_component("fill_value", fill_value, copy=False)
        self._set_component("dtype", dtype, copy=False)
        self._set_component("shape", shape, copy=False)
        self._set_component("units", units, copy=False)
        self._set_component("calendar", calendar, copy=False)

    def __getitem__(self, indices):
        """x.__getitem__(indices) <==> x[indices]

        Returns a numpy array.

        """
        # If 'apply_indices' is False then we can make a filled numpy
        # array that already has the correct shape. Otherwise we need
        # to make one with shape 'self.shape' and then apply the
        # indices to that.
        apply_indices = False

        if indices is Ellipsis:
            array_shape = self.shape
        else:
            array_shape = []
            for i, size, i in zip(indices, self.shape):
                if not isinstance(i, slice):
                    continue

                start, stop, step = i.indices(size)
                a, b = divmod(stop - start, step)
                if b:
                    a += 1

                array_shape.append(a)

            if len(array_shape) != self.ndim:
                apply_indices = True
                array_shape = self.shape

        fill_value = self.get_fill_value()
        if fill_value is np.ma.masked:
            array = np.ma.masked_all(array_shape, dtype=self.dtype)
        elif fill_value is not None:
            array = np.full(
                array_shape, fill_value=fill_value, dtype=self.dtype
            )
        else:
            array = np.empty(array_shape, dtype=self.dtype)

        if apply_indices:
            array = self.get_subspace(array, indices)

        self._set_units()

        return array

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        return f"<CF {self.__class__.__name__}{self.shape}: {self}>"

    def __str__(self):
        """Called by the `str` built-in function.

        x.__str__() <==> str(x)

        """
        fill_value = self.get_fill_value()
        if fill_value is None:
            return "Uninitialised"

        return f"Filled with {fill_value!r}"

    def _set_units(self):
        """The units and calendar properties.

        These are the values set during initialisation, defaulting to
        `None` if either was not set at that time.

        .. versionadded:: TODODASKVER

        :Returns:

            `tuple`
                The units and calendar values, either of which may be
                `None`.

        """
        # TODOCFA: Consider moving _set_units to cfdm.Array, or some
        #          other common ancestor so that this, and other,
        #          subclasses can access it.
        units = self.get_units(False)
        if units is False:
            units = None
            self._set_component("units", units, copy=False)

        calendar = self.get_calendar(False)
        if calendar is False:
            calendar = None
            self._set_component("calendar", calendar, copy=False)

        return units, calendar

    @property
    def dtype(self):
        """Data-type of the data elements."""
        return self._get_component("dtype")

    @property
    def shape(self):
        """Tuple of array dimension sizes."""
        return self._get_component("shape")

    def get_fill_value(self):
        """Return the data array fill value.

        .. versionadded:: TODODASKVER

        :Returns:

                The fill value.

        """
        return self._get_component("fill_value", None)



import random
from string import hexdigits

import cfdm
import dask.array as da
import numpy as np

from ... import Bounds, Coordinate, DomainAncillary
from .netcdfread import NetCDFRead

_cfa_message = (
    "Writing CFA files has been temporarily disabled, "
    "and will return at version 4.0.0. "
    "CFA-0.4 functionality is still available at version 3.13.x."
)


class NetCDFWrite(cfdm.read_write.netcdf.NetCDFWrite):
    """A container for writing Fields to a netCDF dataset."""

    def __new__(cls, *args, **kwargs):
        """Store the NetCDFRead class.

        .. note:: If a child class requires a different NetCDFRead class
        than the one defined here, then it must be redefined in the
        child class.

        """
        instance = super().__new__(cls)
        instance._NetCDFRead = NetCDFRead
        return instance

    def _write_as_cfa(self, cfvar):
        """True if the variable should be written as a CFA variable.

        .. versionadded:: 3.0.0

        """
        if not self.write_vars["cfa"]:
            return False

        data = self.implementation.get_data(cfvar, None)
        if data is None:
            return False

        if data.size == 1:
            return False

        if isinstance(cfvar, (Coordinate, DomainAncillary)):
            return cfvar.ndim > 1

        if isinstance(cfvar, Bounds):
            return cfvar.ndim > 2

        return True

    def _customize_createVariable(self, cfvar, kwargs):
        """Customise keyword arguments for
        `netCDF4.Dataset.createVariable`.

        .. versionadded:: 3.0.0

        :Parameters:

            cfvar: cf instance that contains data

            kwargs: `dict`

        :Returns:

            `dict`
                Dictionary of keyword arguments to be passed to
                `netCDF4.Dataset.createVariable`.

        """
        kwargs = super()._customize_createVariable(cfvar, kwargs)

        if self._write_as_cfa(cfvar):
            raise ValueError(_cfa_message)

            kwargs["dimensions"] = ()
            kwargs["chunksizes"] = None

        return kwargs

    def _write_data(
        self,
        data,
        cfvar,
        ncvar,
        ncdimensions,
        unset_values=(),
        compressed=False,
        attributes={},
    ):
        """Write a Data object.

        .. versionadded:: 3.0.0

        :Parameters:

            data: `Data`

            cfvar: cf instance

            ncvar: `str`

            ncdimensions: `tuple` of `str`

            unset_values: sequence of numbers

        """
        g = self.write_vars

        if self._write_as_cfa(cfvar):
            raise ValueError(_cfa_message)

            self._write_cfa_data(ncvar, ncdimensions, data, cfvar)
            return

        # Still here?
        if compressed:
            # --------------------------------------------------------
            # Write data in its compressed form
            # --------------------------------------------------------
            data = data.source().source()

        # Get the dask array
        dx = da.asanyarray(data)

        # Convert the data type
        new_dtype = g["datatype"].get(dx.dtype)
        if new_dtype is not None:
            dx = dx.astype(new_dtype)

        # VLEN variables can not be assigned to by masked arrays
        # (https://github.com/Unidata/netcdf4-python/pull/465), so
        # fill missing data in string (as opposed to char) data types.
        if g["fmt"] == "NETCDF4" and dx.dtype.kind in "SU":
            dx = dx.map_blocks(
                self._filled_string_array,
                fill_value="",
                meta=np.array((), dx.dtype),
            )

        # Check for out-of-range values
        if g["warn_valid"]:
            dx = dx.map_blocks(
                self._check_valid,
                cfvar=cfvar,
                attributes=attributes,
                meta=np.array((), dx.dtype),
            )

        da.store(dx, g["nc"][ncvar], compute=True, return_stored=False)

    def _write_dimension_coordinate(
        self, f, key, coord, ncdim=None, coordinates=None
    ):
        """Write a coordinate variable and its bound variable to the
        file.

        This also writes a new netCDF dimension to the file and, if
        required, a new netCDF dimension for the bounds.

        .. versionadded:: 3.0.0

        :Parameters:

            f: Field construct

            key: `str`

            coord: Dimension coordinate construct

            ncdim: `str` or `None`
                The name of the netCDF dimension for this dimension
                coordinate construct, including any groups structure. Note
                that the group structure may be different to the
                corodinate variable, and the basename.

                .. versionadded:: 3.6.0

            coordinates: `list`
               This list may get updated in-place.

                .. versionadded:: 3.7.0

        :Returns:

            `str`
                The netCDF name of the dimension coordinate.

        """
        coord = self._change_reference_datetime(coord)

        return super()._write_dimension_coordinate(
            f, key, coord, ncdim=ncdim, coordinates=coordinates
        )

    def _write_scalar_coordinate(
        self, f, key, coord_1d, axis, coordinates, extra=None
    ):
        """Write a scalar coordinate and its bounds to the netCDF file.

        It is assumed that the input coordinate has size 1, but this is
        not checked.

        If an equal scalar coordinate has already been written to the file
        then the input coordinate is not written.

        .. versionadded:: 3.0.0

        :Parameters:

            f: Field construct

            key: `str`
                The coordinate construct key

            coord_1d: Coordinate construct

            axis: `str`
                The field's axis identifier for the scalar coordinate.

            coordinates: `list`

        :Returns:

            coordinates: `list`
                The updated list of netCDF auxiliary coordinate names.

        """
        # Unsafe to set mutable '{}' as default in the func signature.
        if extra is None:  # distinguish from falsy '{}'
            extra = {}

        coord_1d = self._change_reference_datetime(coord_1d)

        return super()._write_scalar_coordinate(
            f, key, coord_1d, axis, coordinates, extra=extra
        )

    def _write_auxiliary_coordinate(self, f, key, coord, coordinates):
        """Write auxiliary coordinates and bounds to the netCDF file.

        If an equal auxiliary coordinate has already been written to the
        file then the input coordinate is not written.

        .. versionadded:: 3.0.0

        :Parameters:

            f: Field construct

            key: `str`

            coord: Coordinate construct

            coordinates: `list`

        :Returns:

            coordinates: `list`
                The list of netCDF auxiliary coordinate names updated in
                place.

        **Examples:**

        >>> coordinates = _write_auxiliary_coordinate(f, 'aux2', coordinates)

        """
        coord = self._change_reference_datetime(coord)

        return super()._write_auxiliary_coordinate(f, key, coord, coordinates)

    def _change_reference_datetime(self, coord):
        """Change the units of a reference date-time value.

        .. versionadded:: 3.0.0

        :Parameters:

            coord: Coordinate instance

        :Returns:

                The coordinate construct with changed units.

        """
        reference_datetime = self.write_vars.get("reference_datetime")
        if not reference_datetime or not coord.Units.isreftime:
            return coord

        coord2 = coord.copy()
        try:
            coord2.reference_datetime = reference_datetime
        except ValueError:
            raise ValueError(
                "Can't override coordinate reference date-time "
                f"{coord.reference_datetime!r} with {reference_datetime!r}"
            )
        else:
            return coord2

    def _write_cfa_data(self, ncvar, ncdimensions, data, cfvar):
        """Write a CFA variable to the netCDF file.

        Any CFA private variables required will be autmatically created
        and written to the file.

        .. versionadded:: 3.0.0

        :Parameters:

            ncvar: `str`
                The netCDF name for the variable.

            ncdimensions: sequence of `str`

            netcdf_attrs: `dict`

            data: `Data`

        :Returns:

            `None`

        """
        raise ValueError(_cfa_message)

    def _random_hex_string(self, size=10):
        """Return a random hexadecimal string with the given number of
        characters.

        .. versionadded:: 3.0.0

        :Parameters:

            size: `int`, optional
                The number of characters in the generated string.

        :Returns:

            `str`
                The hexadecimal string.

        **Examples:**

        >>> _random_hex_string()
        'C3eECbBBcf'
        >>> _random_hex_string(6)
        '7a4acc'

        """
        return "".join(random.choice(hexdigits) for i in range(size))

    def _convert_to_builtin_type(self, x):
        """Convert a non-JSON-encodable object to a JSON-encodable
        built-in type.

        Possible conversions are:

        ==============  =============  ======================================
        Input object    Output object  numpy data types covered
        ==============  =============  ======================================
        numpy.bool_     bool           bool
        numpy.integer   int            int, int8, int16, int32, int64, uint8,
                                       uint16, uint32, uint64
        numpy.floating  float          float, float16, float32, float64
        ==============  =============  ======================================

        .. versionadded:: 3.0.0

        :Parameters:

            x:

        :Returns:

            'int' or `float` or `bool`

        **Examples:**

        >>> type(_convert_to_builtin_type(numpy.bool_(True)))
        bool
        >>> type(_convert_to_builtin_type(numpy.array([1.0])[0]))
        double
        >>> type(_convert_to_builtin_type(numpy.array([2])[0]))
        int

        """
        if isinstance(x, np.bool_):
            return bool(x)

        if isinstance(x, np.integer):
            return int(x)

        if isinstance(x, np.floating):
            return float(x)

        raise TypeError(
            f"{type(x)!r} object can't be converted to a JSON serializable "
            f"type: {x!r}"
        )

    def _check_valid(self, array, cfvar=None, attributes=None):
        """Checks for array values outside of the valid range.

        Specifically, checks array for out-of-range values, as
        defined by the valid_[min|max|range] attributes.

        .. versionadded:: TODODASKVER

        :Parameters:

            array: `numpy.ndarray`
                The array to be checked.

            cfvar: construct
                The CF construct containing the array.

            attributes: `dict`
                The variable's CF properties.

        :Returns:

            `numpy.ndarray`
                The input array, unchanged.

        """
        super()._check_valid(cfvar, array, attributes)
        return array

    def _filled_string_array(self, array, fill_value=""):
        """Fill a string array.

        .. versionadded:: TODODASKVER

        :Parameters:

            array: `numpy.ndarray`
                The `numpy` array with string (byte or unicode) data
                type.

        :Returns:

            `numpy.ndarray`
                The string array array with any missing data replaced
                by the fill value.

        """
        if np.ma.isMA(array):
            return array.filled(fill_value)

        return array


#    def _convert_dtype(self, array, new_dtype=None):
#        """Convert the data type of a numpy array.
#
#        .. versionadded:: TODODASKVER
#
#        :Parameters:
#
#            array: `numpy.ndarray`
#                The `numpy` array
#
#            new_dtype: data-type
#                The new data type.
#
#        :Returns:
#
#            `numpy.ndarray`
#                The array with converted data type.
#
#        """
#        return array.astype(new_dtype)
#



from ....functions import _DEPRECATION_ERROR_ATTRIBUTE
from ..mixin import FileArrayMixin
from .array import Array


class FileArray(FileArrayMixin, Array):
    """Abstract base class for an array stored in a file."""

    def __getitem__(self, indices):
        """Return a subspace of the array.

        x.__getitem__(indices) <==> x[indices]

        Returns a subspace of the array as an independent numpy array.

        """
        raise NotImplementedError(
            f"Must implement {self.__class__.__name__}.__getitem__"
        )  # pragma: no cover

    def __repr__(self):
        """x.__repr__() <==> repr(x)"""
        return f"<CF {self.__class__.__name__}{self.shape}: {self}>"

    def __str__(self):
        """x.__str__() <==> str(x)"""
        return f"{self.get_filename()}, {self.get_address()}"

    @property
    def dtype(self):
        """Data-type of the array."""
        return self._get_component("dtype")

    @property
    def filename(self):
        """The name of the file containing the array.

        Deprecated at version TODODASKVER. Use method `get_filename`
        instead.

        """
        _DEPRECATION_ERROR_ATTRIBUTE(
            self,
            "filename",
            message="Use method 'get_filename' instead.",
            version="TODODASKVER",
            removed_at="5.0.0",
        )  # pragma: no cover

    @property
    def shape(self):
        """Shape of the array."""
        return self._get_component("shape")

    def close(self):
        """Close the dataset containing the data."""
        raise NotImplementedError(
            f"Must implement {self.__class__.__name__}.close"
        )  # pragma: no cover

    def get_address(self):
        """The address in the file of the variable.

        .. versionadded:: TODODASKVER

        :Returns:

            `str` or `None`
                The address, or `None` if there isn't one.

        """
        raise NotImplementedError(
            f"Must implement {self.__class__.__name__}.get_address"
        )  # pragma: no cover

    def get_filename(self):
        """Return the name of the file containing the array.

        :Returns:

            `str` or `None`
                The filename, or `None` if there isn't one.

        **Examples**

        >>> a.get_filename()
        'file.nc'

        """
        return self._get_component("filename", None)

    def open(self):
        """Returns an open dataset containing the data array."""
        raise NotImplementedError(
            f"Must implement {self.__class__.__name__}.open"
        )  # pragma: no cover



from ..array.umarray import UMArray
from .abstract import FragmentArray


class UMFragmentArray(FragmentArray):
    """A CFA fragment array stored in a UM or PP file.

    .. versionadded:: TODODASKVER

    """

    def __init__(
        self,
        filename=None,
        address=None,
        dtype=None,
        shape=None,
        aggregated_units=False,
        aggregated_calendar=False,
        units=False,
        calendar=False,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            filename: `str`
                The name of the UM or PP file containing the fragment.

            address: `int`, optional
                The start word in the file of the header.

            dtype: `numpy.dtype`
                The data type of the aggregated array. May be `None`
                if the numpy data-type is not known (which can be the
                case for netCDF string types, for example). This may
                differ from the data type of the netCDF fragment
                variable.

            shape: `tuple`
                The shape of the fragment within the aggregated
                array. This may differ from the shape of the netCDF
                fragment variable in that the latter may have fewer
                size 1 dimensions.

            units: `str` or `None`, optional
                The units of the fragment data. Set to `None` to
                indicate that there are no units. If unset then the
                units will be set during the first `__getitem__` call.

            calendar: `str` or `None`, optional
                The calendar of the fragment data. Set to `None` to
                indicate the CF default calendar, if applicable. If
                unset then the calendar will be set during the first
                `__getitem__` call.

            {{aggregated_units: `str` or `None`, optional}}

            {{aggregated_calendar: `str` or `None`, optional}}

            source: optional
                Initialise the array from the given object.

                {{init source}}

            {{deep copy}}

        """
        if source is not None:
            super().__init__(source=source, copy=copy)
            return

        array = UMArray(
            filename=filename,
            header_offset=address,
            dtype=dtype,
            shape=shape,
            units=units,
            calendar=calendar,
            copy=False,
        )

        super().__init__(
            filename=filename,
            address=address,
            dtype=dtype,
            shape=shape,
            aggregated_units=aggregated_units,
            aggregated_calendar=aggregated_calendar,
            array=array,
            source=source,
            copy=False,
        )



import datetime
import faulthandler
import os
import unittest
import warnings

faulthandler.enable()  # to debug seg faults and timeouts

import cf

# To facilitate the testing of logging outputs (see test_aggregate_verbosity)
log_name = __name__
logger = cf.logging.getLogger(log_name)


class aggregateTest(unittest.TestCase):
    filename = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "test_file.nc"
    )
    file = os.path.join(os.path.dirname(os.path.abspath(__file__)), "file.nc")
    file2 = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "file2.nc"
    )

    chunk_sizes = (100000, 300, 34)
    original_chunksize = cf.chunksize()

    def test_basic_aggregate(self):
        for chunksize in self.chunk_sizes:
            cf.chunksize(chunksize)

            f = cf.read(self.filename, squeeze=True)[0]

            g = cf.FieldList(f[0])
            g.append(f[1:3])
            g.append(f[3])
            g[-1].flip(0, inplace=True)
            g.append(f[4:7])
            g[-1].flip(0, inplace=True)
            g.extend([f[i] for i in range(7, f.shape[0])])

            g0 = g.copy()
            self.assertTrue(g.equals(g0, verbose=-1), "g != g0")

            with warnings.catch_warnings():
                # Suppress noise throughout the test fixture from:
                #
                #   ~/cf-python/cf/__init__.py:1459: FutureWarning: elementwise
                #   comparison failed; returning scalar instead, but in the
                #   future will perform elementwise comparison
                #
                # TODO: it is not clear where the above emerges from, e.g.
                # since __init__ file ref'd does not have that many lines.
                # It seems like this warning arises from NumPy comparisons
                # done at some point in (only) some aggregate calls (see e.g:
                # https://github.com/numpy/numpy/issues/6784).
                warnings.filterwarnings("ignore", category=FutureWarning)
                h = cf.aggregate(g, verbose=2)

            self.assertEqual(len(h), 1)

            self.assertEqual(
                h[0].shape,
                (10, 9),
                "h[0].shape = " + repr(h[0].shape) + " != (10, 9)",
            )

            self.assertTrue(
                g.equals(g0, verbose=2), "g != itself after aggregation"
            )

            self.assertTrue(h[0].equals(f, verbose=2), "h[0] != f")

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=FutureWarning)
                i = cf.aggregate(g, verbose=2)

            self.assertTrue(
                i.equals(h, verbose=2), "The second aggregation != the first"
            )

            self.assertTrue(
                g.equals(g0, verbose=2),
                "g != itself after the second aggregation",
            )

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=FutureWarning)
                i = cf.aggregate(g, verbose=2, axes="grid_latitude")

            self.assertTrue(
                i.equals(h, verbose=2), "The third aggregation != the first"
            )

            self.assertTrue(
                g.equals(g0, verbose=2),
                "g !=itself after the third aggregation",
            )

            self.assertEqual(
                i[0].shape, (10, 9), "i[0].shape is " + repr(i[0].shape)
            )

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=FutureWarning)
                i = cf.aggregate(
                    g,
                    verbose=2,
                    axes="grid_latitude",
                    donotchecknonaggregatingaxes=1,
                )

            self.assertTrue(
                i.equals(h, verbose=2), "The fourth aggregation != the first"
            )

            self.assertTrue(
                g.equals(g0, verbose=2),
                "g != itself after the fourth aggregation",
            )

            self.assertEqual(
                i[0].shape, (10, 9), "i[0].shape is " + repr(i[0].shape)
            )

            q, t = cf.read(self.file)
            c = cf.read(self.file2)[0]

            d = cf.aggregate([c, t], verbose=1, relaxed_identities=True)
            e = cf.aggregate([t, c], verbose=1, relaxed_identities=True)

            self.assertEqual(len(d), 1)
            self.assertEqual(len(e), 1)
            self.assertEqual(d[0].shape, (3,) + t.shape)
            self.assertTrue(d[0].equals(e[0], verbose=2))

            x = cf.read(["file.nc", "file2.nc"], aggregate=False)
            self.assertEqual(len(x), 3)

            x = cf.read(
                ["file.nc", "file2.nc"], aggregate={"relaxed_identities": True}
            )
            self.assertEqual(len(x), 2)

            del t.standard_name
            del c.standard_name
            x = cf.aggregate([c, t], verbose=1)
            self.assertEqual(len(x), 2)

            t.long_name = "qwerty"
            c.long_name = "qwerty"
            x = cf.aggregate([c, t], field_identity="long_name")
            self.assertEqual(len(x), 1)

        cf.chunksize(self.original_chunksize)

    def test_aggregate_exist_equal_ignore_opts(self):
        # TODO: extend the option-checking coverage so all options and all
        # reasonable combinations of them are tested. For now, this is
        # testing options that previously errored due to a bug.
        for chunksize in self.chunk_sizes:
            cf.chunksize(chunksize)

            f = cf.read(self.filename, squeeze=True)[0]

            # Use f as-is: simple test that aggregate works and does not
            # change anything with the given options:
            g = cf.aggregate(f, exist_all=True)[0]
            self.assertEqual(g, f)
            h = cf.aggregate(f, equal_all=True)[0]
            self.assertEqual(h, f)

            with self.assertRaises(ValueError):  # contradictory options
                cf.aggregate(f, exist_all=True, equal_all=True)

        cf.chunksize(self.original_chunksize)

    def test_aggregate_verbosity(self):
        f0 = cf.example_field(0)
        f1 = cf.example_field(1)

        detail_header = "DETAIL:cf.aggregate:STRUCTURAL SIGNATURE:"
        debug_header = "DEBUG:cf.aggregate:COMPLETE AGGREGATION METADATA:"

        # 'DEBUG' (-1) verbosity should output both log message headers...
        with self.assertLogs(level="NOTSET") as catch:
            cf.aggregate([f0, f1], verbose=-1)
            for header in (detail_header, debug_header):
                self.assertTrue(
                    any(
                        log_item.startswith(header)
                        for log_item in catch.output
                    ),
                    f"No log entry begins with '{header}'",
                )

        # ...but with 'DETAIL' (3), should get only the detail-level one.
        with self.assertLogs(level="NOTSET") as catch:
            cf.aggregate([f0, f1], verbose=3)
            self.assertTrue(
                any(
                    log_item.startswith(detail_header)
                    for log_item in catch.output
                ),
                f"No log entry begins with '{detail_header}'",
            )
            self.assertFalse(
                any(
                    log_item.startswith(debug_header)
                    for log_item in catch.output
                ),
                f"A log entry begins with '{debug_header}' but should not",
            )

        # and neither should emerge at the 'WARNING' (1) level.
        with self.assertLogs(level="NOTSET") as catch:
            logger.warning(
                "Dummy message to log something at warning level so that "
                "'assertLog' does not error when no logs messages emerge."
            )
            # Note: can use assertNoLogs in Python 3.10 to avoid this, see:
            # https://bugs.python.org/issue39385

            cf.aggregate([f0, f1], verbose=1)
            for header in (detail_header, debug_header):
                self.assertFalse(
                    any(
                        log_item.startswith(header)
                        for log_item in catch.output
                    ),
                    f"A log entry begins with '{header}' but should not",
                )

    def test_aggregate_bad_units(self):
        f = cf.read(self.filename, squeeze=True)[0]

        g = cf.FieldList(f[0])
        g.append(f[1:])

        h = cf.aggregate(g)
        self.assertEqual(len(h), 1)

        g[0].override_units(cf.Units("apples!"), inplace=True)
        g[1].override_units(cf.Units("oranges!"), inplace=True)

        h = cf.aggregate(g)
        self.assertEqual(len(h), 2)

    def test_aggregate_domain(self):
        f = cf.example_field(0)
        g = f[0:3].domain
        h = f[3:].domain

        x = cf.aggregate([g, h])

        self.assertEqual(len(x), 1, x)

    def test_aggregate_dimension(self):
        """Test the promotion of property to axis."""
        f = cf.example_field(0)
        g = f.copy()

        f.set_property("sim", "r1i1p1f1")
        g.set_property("sim", "r2i1p1f1")

        self.assertFalse(len(f.auxiliary_coordinates()))

        a = cf.aggregate([f, g], dimension="sim")
        self.assertEqual(len(a), 1)

        a = a[0]
        self.assertEqual(len(a.auxiliary_coordinates()), 1)

    def test_aggregate_keyword_consistency(self):
        """Test acceptable keyword combinations."""
        f = cf.example_field(0)
        a = cf.aggregate(
            [f[:2], f[2:]], relaxed_identities=True, ncvar_identities=True
        )
        self.assertEqual(len(a), 1)

    def test_aggregate_equal_equal_all(self):
        f = cf.example_field(0)
        f.set_property("foo", "bar")
        a, b, c, d = f[0], f[1], f[2:4], f[4]

        c.set_property("foo", "baz")
        d.set_property("foo", "baz")

        g = cf.aggregate([a, b, c, d])
        self.assertEqual(len(g), 1)

        g = cf.aggregate([a, b, c, d], equal=["foo"])
        self.assertEqual(len(g), 2)

        g = cf.aggregate([a, b, c, d], equal_all=True)
        self.assertEqual(len(g), 2)

        d.del_property("foo")

        g = cf.aggregate([a, b, c, d])
        self.assertEqual(len(g), 1)

        g = cf.aggregate([a, b, c, d], equal=["foo"])
        self.assertEqual(len(g), 3)

        g = cf.aggregate([a, b, c, d], equal_all=True)
        self.assertEqual(len(g), 3)

    def test_aggregate_exist_exist_all(self):
        f = cf.example_field(0)
        f.set_property("foo", "bar")
        a, b, c, d = f[0], f[1], f[2:4], f[4]

        c.set_property("foo", "baz")
        d.set_property("foo", "baz")

        g = cf.aggregate([a, b, c, d])
        self.assertEqual(len(g), 1)

        g = cf.aggregate([a, b, c, d], exist=["foo"])
        self.assertEqual(len(g), 1)

        g = cf.aggregate([a, b, c, d], exist_all=True)
        self.assertEqual(len(g), 1)

        d.del_property("foo")

        g = cf.aggregate([a, b, c, d])
        self.assertEqual(len(g), 1)

        g = cf.aggregate([a, b, c, d], exist=["foo"])
        self.assertEqual(len(g), 2)

        g = cf.aggregate([a, b, c, d], exist_all=True)
        self.assertEqual(len(g), 2)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import atexit
import datetime
import faulthandler
import inspect
import os
import shutil
import subprocess
import tempfile
import unittest

import numpy

faulthandler.enable()  # to debug seg faults and timeouts

import cf

n_tmpfiles = 8
tmpfiles = [
    tempfile.mkstemp("_test_read_write.nc", dir=os.getcwd())[1]
    for i in range(n_tmpfiles)
]
(
    tmpfile,
    tmpfileh,
    tmpfileh2,
    tmpfilec,
    tmpfilec2,
    tmpfile0,
    tmpfile1,
    tmpfile2,
) = tmpfiles


def _remove_tmpfiles():
    """Try to remove defined temporary files by deleting their paths."""
    for f in tmpfiles:
        try:
            os.remove(f)
        except OSError:
            pass


atexit.register(_remove_tmpfiles)

TEST_DASKIFIED_ONLY = True


class read_writeTest(unittest.TestCase):
    filename = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "test_file.nc"
    )

    broken_bounds = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "broken_bounds.cdl"
    )

    string_filename = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "string_char.nc"
    )

    chunk_sizes = (100000, 300)

    f0 = cf.example_field(0)
    f1 = cf.example_field(1)

    netcdf3_fmts = [
        "NETCDF3_CLASSIC",
        "NETCDF3_64BIT",
        "NETCDF3_64BIT_OFFSET",
        "NETCDF3_64BIT_DATA",
    ]
    netcdf4_fmts = ["NETCDF4", "NETCDF4_CLASSIC"]
    netcdf_fmts = netcdf3_fmts + netcdf4_fmts

    #    @unittest.skipIf(TEST_DASKIFIED_ONLY, "KeyError: 'q'")
    def test_write_filename(self):
        f = self.f0
        a = f.array

        cf.write(f, tmpfile)
        g = cf.read(tmpfile)

        self.assertTrue((a == g[0].array).all())

    def test_read_mask(self):
        f = self.f0.copy()

        N = f.size

        f.data[1, 1] = cf.masked
        f.data[2, 2] = cf.masked

        f.del_property("_FillValue", None)
        f.del_property("missing_value", None)

        cf.write(f, tmpfile)

        g = cf.read(tmpfile)[0]
        self.assertEqual(numpy.ma.count(g.data.array), N - 2)

        g = cf.read(tmpfile, mask=False)[0]
        self.assertEqual(numpy.ma.count(g.data.array), N)

        g.apply_masking(inplace=True)
        self.assertEqual(numpy.ma.count(g.data.array), N - 2)

        f.set_property("_FillValue", 999)
        f.set_property("missing_value", -111)
        cf.write(f, tmpfile)

        g = cf.read(tmpfile)[0]
        self.assertEqual(numpy.ma.count(g.data.array), N - 2)

        g = cf.read(tmpfile, mask=False)[0]
        self.assertEqual(numpy.ma.count(g.data.array), N)

        g.apply_masking(inplace=True)
        self.assertEqual(numpy.ma.count(g.data.array), N - 2)

    def test_read_directory(self):
        pwd = os.getcwd() + "/"

        dir = "dir_" + inspect.stack()[0][3]

        try:
            os.mkdir(dir)
        except FileExistsError:
            pass
        except Exception:
            raise ValueError(f"Can not mkdir {pwd}{dir}")

        f = "test_file2.nc"
        try:
            os.symlink(pwd + f, pwd + dir + "/" + f)
        except FileExistsError:
            pass

        subdir = dir + "/subdir"
        try:
            os.mkdir(subdir)
        except FileExistsError:
            pass
        except Exception:
            raise ValueError(f"Can not mkdir {pwd}{subdir}")

        for f in ("test_file3.nc", "test_file.nc"):
            try:
                os.symlink(pwd + f, pwd + subdir + "/" + f)
            except FileExistsError:
                pass

        f = cf.read(dir, aggregate=False)
        self.assertEqual(len(f), 1, f)

        f = cf.read(dir, recursive=True, aggregate=False)
        self.assertEqual(len(f), 3)

        f = cf.read([dir, subdir], aggregate=False)
        self.assertEqual(len(f), 3)

        f = cf.read([subdir, dir], aggregate=False)
        self.assertEqual(len(f), 3)

        f = cf.read([dir, subdir], recursive=True, aggregate=False)
        self.assertEqual(len(f), 5)

        f = cf.read(subdir, aggregate=False)
        self.assertEqual(len(f), 2)

        f = cf.read(subdir, recursive=True, aggregate=False)
        self.assertEqual(len(f), 2)

        shutil.rmtree(dir)

    def test_read_select(self):
        # select on field list
        f = cf.read(self.filename, select="eastward_wind")
        g = cf.read(self.filename)
        self.assertTrue(f.equals(g, verbose=2), "Bad read with select keyword")

    def test_read_squeeze(self):
        # select on field list
        cf.read(self.filename, squeeze=True)
        cf.read(self.filename, unsqueeze=True)
        with self.assertRaises(Exception):
            cf.read(self.filename, unsqueeze=True, squeeze=True)

    def test_read_aggregate(self):
        cf.read(self.filename, aggregate=True)
        cf.read(self.filename, aggregate=False)
        cf.read(self.filename, aggregate={})

    def test_read_extra(self):
        # Test field keyword of cf.read
        filename = self.filename

        f = cf.read(filename)
        self.assertEqual(len(f), 1, "\n" + str(f))

        f = cf.read(filename, extra=["auxiliary_coordinate"])
        self.assertEqual(len(f), 4, "\n" + str(f))

        f = cf.read(filename, extra="cell_measure")
        self.assertEqual(len(f), 2, "\n" + str(f))

        f = cf.read(filename, extra=["field_ancillary"])
        self.assertEqual(len(f), 5, "\n" + str(f))

        f = cf.read(filename, extra="domain_ancillary", verbose=0)
        self.assertEqual(len(f), 4, "\n" + str(f))

        f = cf.read(
            filename, extra=["field_ancillary", "auxiliary_coordinate"]
        )
        self.assertEqual(len(f), 8, "\n" + str(f))

        self.assertEqual(
            len(
                cf.read(
                    filename,
                    extra=["domain_ancillary", "auxiliary_coordinate"],
                )
            ),
            7,
        )
        f = cf.read(
            filename,
            extra=["domain_ancillary", "cell_measure", "auxiliary_coordinate"],
        )
        self.assertEqual(len(f), 8, "\n" + str(f))

        f = cf.read(
            filename,
            extra=(
                "field_ancillary",
                "dimension_coordinate",
                "cell_measure",
                "auxiliary_coordinate",
                "domain_ancillary",
            ),
        )
        self.assertEqual(len(f), 15, "\n" + str(f))

    def test_read_write_format(self):
        cf.write(self.f1, tmpfile)
        f = cf.read(tmpfile)[0]

        # TODO: reinstate "CFA" at version > 3.14
        for fmt in self.netcdf_fmts:  # + ["CFA"]:
            cf.write(f, tmpfile2, fmt=fmt)
            g = cf.read(tmpfile2, verbose=0)
            self.assertEqual(len(g), 1)
            g = g[0]

            self.assertTrue(
                f.equals(g, verbose=1),
                f"Bad read/write of format {fmt!r}",
            )

    def test_write_netcdf_mode(self):
        """Test the `mode` parameter to `write`, notably append mode."""
        g = cf.read(self.filename)  # note 'g' has one field

        # Test special case #1: attempt to append fields with groups
        # (other than 'root') which should be forbidden. Using fmt="NETCDF4"
        # since it is the only format where groups are allowed.
        #
        # Note: this is not the most natural test to do first, but putting
        # it before the rest reduces spurious seg faults for me, so...
        g[0].nc_set_variable_groups(["forecast", "model"])
        cf.write(g, tmpfile, fmt="NETCDF4", mode="w")  # 1. overwrite to wipe
        f = cf.read(tmpfile)
        with self.assertRaises(ValueError):
            cf.write(g[0], tmpfile, fmt="NETCDF4", mode="a")

        # Test special case #2: attempt to append fields with contradictory
        # featureType to the original file:
        g[0].nc_clear_variable_groups()
        g[0].nc_set_global_attribute("featureType", "profile")
        cf.write(
            g,
            tmpfile,
            fmt="NETCDF4",
            mode="w",
            global_attributes=("featureType", "profile"),
        )  # 1. overwrite to wipe
        h = cf.example_field(3)
        h.nc_set_global_attribute("featureType", "timeSeries")
        with self.assertRaises(ValueError):
            cf.write(h, tmpfile, fmt="NETCDF4", mode="a")
        # Now remove featureType attribute for subsquent tests:
        g_attrs = g[0].nc_clear_global_attributes()
        del g_attrs["featureType"]
        g[0].nc_set_global_attributes(g_attrs)

        # Set a non-trivial (i.e. not only 'Conventions') global attribute to
        # make the global attribute testing more robust:
        add_global_attr = ["remark", "A global comment."]
        original_global_attrs = g[0].nc_global_attributes()
        original_global_attrs[add_global_attr[0]] = None  # -> None on fields
        g[0].nc_set_global_attribute(*add_global_attr)

        # First test a bad mode value:
        with self.assertRaises(ValueError):
            cf.write(g[0], tmpfile, mode="g")

        g_copy = g.copy()

        for fmt in self.netcdf_fmts:  # test over all netCDF 3 and 4 formats
            # Other tests cover write as default mode (i.e. test with no mode
            # argument); here test explicit provision of 'w' as argument:
            cf.write(
                g,
                tmpfile,
                fmt=fmt,
                mode="w",
                global_attributes=add_global_attr,
            )
            f = cf.read(tmpfile)

            new_length = 1  # since 1 == len(g)
            self.assertEqual(len(f), new_length)
            # Ignore as 'remark' should be 'None' on the field as tested below
            self.assertTrue(f[0].equals(g[0], ignore_properties=["remark"]))
            self.assertEqual(
                f[0].nc_global_attributes(), original_global_attrs
            )

            # Main aspect of this test: testing the append mode ('a'): now
            # append all other example fields, to check a diverse variety.
            for ex_field_n, ex_field in enumerate(cf.example_fields()):
                # Note: after Issue #141, this skip can be removed.
                if ex_field_n == 1:
                    continue

                # Skip since "RuntimeError: Can't create variable in
                # NETCDF4_CLASSIC file from (2)  (NetCDF: Attempting netcdf-4
                # operation on strict nc3 netcdf-4 file)" i.e. not possible.
                if fmt == "NETCDF4_CLASSIC" and ex_field_n in (6, 7):
                    continue

                cf.write(ex_field, tmpfile, fmt=fmt, mode="a")
                f = cf.read(tmpfile)

                if ex_field_n == 5:  # another special case
                    # The n=2 and n=5 example fields for cf-python aggregate
                    # down to one field, e.g. for b as n=2 and c as n=5:
                    #   >>> c.equals(b, verbose=-1)
                    #   Data: Different shapes: (118, 5, 8) != (36, 5, 8)
                    #   Field: Different data
                    #   False
                    #   >>> a = cf.aggregate([b, c])
                    #   >>> a
                    #   [<CF Field: air_potential_temperature(
                    #    time(154), latitude(5), longitude(8)) K>]
                    #
                    # therefore need to check FL length hasn't changed and
                    # (further below) that n=2,5 aggregated field is present.
                    pass  # i.e. new_length should remain the same as before
                else:
                    new_length += 1  # should be exactly one more field now
                self.assertEqual(len(f), new_length)

                if ex_field_n == 5:
                    ex_n2_and_n5_aggregated = cf.aggregate(
                        [cf.example_field(2), cf.example_field(5)]
                    )[0]
                    self.assertTrue(
                        any(
                            [
                                ex_n2_and_n5_aggregated.equals(
                                    file_field,
                                    ignore_properties=[
                                        "comment",
                                        "featureType",
                                        "remark",
                                    ],
                                )
                                for file_field in f
                            ]
                        )
                    )
                else:
                    # Can't guarantee order of fields created during append op.
                    # so check new field is *somewhere* in read-in fieldlist
                    self.assertTrue(
                        any(
                            [
                                ex_field.equals(
                                    file_field,
                                    ignore_properties=[
                                        "comment",
                                        "featureType",
                                        "remark",
                                    ],
                                )
                                for file_field in f
                            ]
                        )
                    )
                for file_field in f:
                    self.assertEqual(
                        file_field.nc_global_attributes(),
                        original_global_attrs,
                    )

            # Now do the same test, but appending all of the example fields in
            # one operation rather than one at a time, to check that it works.
            cf.write(g, tmpfile, fmt=fmt, mode="w")  # 1. overwrite to wipe
            append_ex_fields = cf.example_fields()
            del append_ex_fields[1]  # note: can remove after Issue #141 closed
            if fmt in "NETCDF4_CLASSIC":
                # Remove n=6 and =7 for reasons as given above (del => minus 1)
                append_ex_fields = append_ex_fields[:5]

            # Equals len(append_ex_fields), + 1 [for original 'g'] and -1 [for
            # field n=5 which aggregates to one with n=2] => + 1 - 1 = + 0:
            overall_length = len(append_ex_fields)
            cf.write(
                append_ex_fields, tmpfile, fmt=fmt, mode="a"
            )  # 2. now append
            f = cf.read(tmpfile)
            self.assertEqual(len(f), overall_length)

            # Also test the mode="r+" alias for mode="a".
            cf.write(g, tmpfile, fmt=fmt, mode="w")  # 1. overwrite to wipe
            cf.write(
                append_ex_fields, tmpfile, fmt=fmt, mode="r+"
            )  # 2. now append
            f = cf.read(tmpfile)
            self.assertEqual(len(f), overall_length)

            # The appended fields themselves are now known to be correct,
            # but we also need to check that any coordinates that are
            # equal across different fields have been shared in the
            # source netCDF, rather than written in separately.
            #
            # Note that the coordinates that are shared across the set of
            # all example fields plus the field 'g' from the contents of
            # the original file (self.filename) are as follows:
            #
            # 1. Example fields n=0 and n=1 share:
            #    <DimensionCoordinate: time(1) days since 2018-12-01 >
            # 2. Example fields n=0, n=2 and n=5 share:
            #    <DimensionCoordinate: latitude(5) degrees_north> and
            #    <DimensionCoordinate: longitude(8) degrees_east>
            # 3. Example fields n=2 and n=5 share:
            #    <DimensionCoordinate: air_pressure(1) hPa>
            # 4. The original file field ('g') and example field n=1 share:
            #    <AuxiliaryCoordinate: latitude(10, 9) degrees_N>,
            #    <AuxiliaryCoordinate: longitude(9, 10) degrees_E>,
            #    <Dimension...: atmosphere_hybrid_height_coordinate(1) >,
            #    <DimensionCoordinate: grid_latitude(10) degrees>,
            #    <DimensionCoordinate: grid_longitude(9) degrees> and
            #    <DimensionCoordinate: time(1) days since 2018-12-01 >
            #
            # Therefore we check all of those coordinates for singularity,
            # i.e. the same underlying netCDF variables, in turn.

            # But first, since the order of the fields appended isn't
            # guaranteed, we must find the mapping of the example fields to
            # their position in the read-in FieldList.
            f = cf.read(tmpfile)
            # Element at index N gives position of example field n=N in file
            file_field_order = []
            for ex_field in cf.example_fields():
                position = [
                    f.index(file_field)
                    for file_field in f
                    if ex_field.equals(
                        file_field,
                        ignore_properties=["comment", "featureType", "remark"],
                    )
                ]
                if not position:
                    position = [None]  # to record skipped example fields
                file_field_order.append(position[0])

            equal_coors = {
                ((0, "dimensioncoordinate2"), (1, "dimensioncoordinate3")),
                ((0, "dimensioncoordinate0"), (2, "dimensioncoordinate1")),
                ((0, "dimensioncoordinate1"), (2, "dimensioncoordinate2")),
                ((0, "dimensioncoordinate0"), (5, "dimensioncoordinate1")),
                ((0, "dimensioncoordinate1"), (5, "dimensioncoordinate2")),
                ((2, "dimensioncoordinate3"), (5, "dimensioncoordinate3")),
            }
            for coor_1, coor_2 in equal_coors:
                ex_field_1_position, c_1 = coor_1
                ex_field_2_position, c_2 = coor_2
                # Now map the appropriate example field to the file FieldList
                f_1 = file_field_order[ex_field_1_position]
                f_2 = file_field_order[ex_field_2_position]
                # None for fields skipped in test, distinguish from falsy 0
                if f_1 is None or f_2 is None:
                    continue
                self.assertEqual(
                    f[f_1]
                    .constructs()
                    .filter_by_identity(c_1)
                    .value()
                    .nc_get_variable(),
                    f[f_2]
                    .constructs()
                    .filter_by_identity(c_2)
                    .value()
                    .nc_get_variable(),
                )

            # Note: after Issue #141, the block below should be un-commented.
            #
            # The original file field 'g' must be at the remaining position:
            # rem_position = list(set(
            #     range(len(f))).difference(set(file_field_order)))[0]
            # # In the final cases, it is easier to remove the one differing
            # # coordinate to get the equal coordinates that should be shared:
            # original_field_coors = dict(f[rem_position].coordinates())
            # ex_field_1_coors = dict(f[file_field_order[1]].coordinates())
            # for orig_coor, ex_1_coor in zip(
            #         original_field_coors.values(), ex_field_1_coors.values()):
            #     # The 'auxiliarycoordinate2' construct differs for both, so
            #     # skip that but otherwise the two fields have the same coors:
            #     if orig_coor.identity == "auxiliarycoordinate2":
            #         continue
            #     self.assertEqual(
            #         orig_coor.nc_get_variable(),
            #         ex_1_coor.nc_get_variable(),
            #     )

            # Check behaviour when append identical fields, as an edge case:
            cf.write(g, tmpfile, fmt=fmt, mode="w")  # 1. overwrite to wipe
            cf.write(g_copy, tmpfile, fmt=fmt, mode="a")  # 2. now append
            f = cf.read(tmpfile)
            self.assertEqual(len(f), 2 * len(g))
            self.assertTrue(
                any(
                    [
                        file_field.equals(g[0], ignore_properties=["remark"])
                        for file_field in f
                    ]
                )
            )
            self.assertEqual(
                f[0].nc_global_attributes(), original_global_attrs
            )

    def test_read_write_netCDF4_compress_shuffle(self):
        f = cf.read(self.filename)[0]
        # TODO: reinstate "CFA4" at version > 3.14
        for fmt in ("NETCDF4", "NETCDF4_CLASSIC"):  # , "CFA4"):
            cf.write(f, tmpfile, fmt=fmt, compress=1, shuffle=True)
            g = cf.read(tmpfile)[0]
            self.assertTrue(
                f.equals(g, verbose=2),
                f"Bad read/write with lossless compression: {fmt}",
            )

    def test_write_datatype(self):
        f = cf.read(self.filename)[0]
        self.assertEqual(f.dtype, numpy.dtype(float))
        cf.write(
            f,
            tmpfile,
            fmt="NETCDF4",
            datatype={numpy.dtype(float): numpy.dtype("float32")},
        )
        g = cf.read(tmpfile)[0]
        self.assertEqual(
            g.dtype,
            numpy.dtype("float32"),
            "datatype read in is " + str(g.dtype),
        )

        # Keyword single
        f = cf.read(self.filename)[0]
        self.assertEqual(f.dtype, numpy.dtype(float))
        cf.write(f, tmpfile, fmt="NETCDF4", single=True)
        g = cf.read(tmpfile)[0]
        self.assertEqual(
            g.dtype,
            numpy.dtype("float32"),
            "datatype read in is " + str(g.dtype),
        )

        # Keyword double
        f = g
        self.assertEqual(f.dtype, numpy.dtype("float32"))
        cf.write(f, tmpfile2, fmt="NETCDF4", double=True)
        g = cf.read(tmpfile2)[0]
        self.assertEqual(
            g.dtype, numpy.dtype(float), "datatype read in is " + str(g.dtype)
        )

        for single in (True, False):
            for double in (True, False):
                with self.assertRaises(Exception):
                    cf.write(g, double=double, single=single)

        datatype = {numpy.dtype(float): numpy.dtype("float32")}
        with self.assertRaises(Exception):
            cf.write(g, datatype=datatype, single=True)

        with self.assertRaises(Exception):
            cf.write(g, datatype=datatype, double=True)

    def test_write_reference_datetime(self):
        for reference_datetime in ("1751-2-3", "1492-12-30"):
            cf.write(self.f0, tmpfile, reference_datetime=reference_datetime)

            g = cf.read(tmpfile)[0]

            t = g.dimension_coordinate("T")
            self.assertEqual(
                t.Units,
                cf.Units("days since " + reference_datetime),
                f"Units written were {t.Units.reftime!r} not "
                f"{reference_datetime!r}",
            )

    def test_read_write_unlimited(self):
        for fmt in ("NETCDF4", "NETCDF3_CLASSIC"):
            f = self.f1.copy()
            domain_axes = f.domain_axes()

            domain_axes["domainaxis0"].nc_set_unlimited(True)
            cf.write(f, tmpfile, fmt=fmt)

            f = cf.read(tmpfile)[0]
            domain_axes = f.domain_axes()
            self.assertTrue(domain_axes["domainaxis0"].nc_is_unlimited())

        fmt = "NETCDF4"
        f = self.f1.copy()
        domain_axes = f.domain_axes()
        domain_axes["domainaxis0"].nc_set_unlimited(True)
        domain_axes["domainaxis2"].nc_set_unlimited(True)
        cf.write(f, tmpfile, fmt=fmt)

        f = cf.read(tmpfile)[0]
        domain_axes = f.domain_axes()
        self.assertTrue(domain_axes["domainaxis0"].nc_is_unlimited())
        self.assertTrue(domain_axes["domainaxis2"].nc_is_unlimited())

    def test_read_CDL(self):
        subprocess.run(
            " ".join(["ncdump", self.filename, ">", tmpfile]),
            shell=True,
            check=True,
        )

        # For the cases of '-h' and '-c', i.e. only header info or coordinates,
        # notably no data, take two cases each: one where there is sufficient
        # info from the metadata to map to fields, and one where there isn't:
        #     1. Sufficient metadata, so should be read-in successfully
        subprocess.run(
            " ".join(["ncdump", "-h", self.filename, ">", tmpfileh]),
            shell=True,
            check=True,
        )
        subprocess.run(
            " ".join(["ncdump", "-c", self.filename, ">", tmpfilec]),
            shell=True,
            check=True,
        )

        #     2. Insufficient metadata, so should error with a message as such
        geometry_1_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "geometry_1.nc"
        )
        subprocess.run(
            " ".join(["ncdump", "-h", geometry_1_file, ">", tmpfileh2]),
            shell=True,
            check=True,
        )
        subprocess.run(
            " ".join(["ncdump", "-c", geometry_1_file, ">", tmpfilec2]),
            shell=True,
            check=True,
        )

        f0 = cf.read(self.filename)[0]

        # Case (1) as above, so read in and check the fields are as should be
        f = cf.read(tmpfile)[0]
        cf.read(tmpfileh)[0]
        c = cf.read(tmpfilec)[0]

        # Case (2) as above, so the right error should be raised on read
        with self.assertRaises(ValueError):
            cf.read(tmpfileh2)[0]

        with self.assertRaises(ValueError):
            cf.read(tmpfilec2)[0]

        self.assertTrue(f0.equals(f, verbose=2))

        self.assertTrue(
            f.construct("grid_latitude").equals(
                c.construct("grid_latitude"), verbose=2
            )
        )
        self.assertTrue(
            f0.construct("grid_latitude").equals(
                c.construct("grid_latitude"), verbose=2
            )
        )

        with self.assertRaises(Exception):
            cf.read("test_read_write.py")

    def test_read_cdl_string(self):
        """Test the `cdl_string` keyword of the `read` function."""
        # Test CDL in full, header-only and coordinate-only type:
        tempfile_to_option_mapping = {
            tmpfile: None,
            tmpfileh: "-h",
            tmpfilec: "-c",
        }

        for tempf, option in tempfile_to_option_mapping.items():
            # Set up the CDL string to test...
            command_to_run = ["ncdump", self.filename, ">", tempf]
            if option:
                command_to_run.insert(1, option)
            subprocess.run(" ".join(command_to_run), shell=True, check=True)
            with open(tempf, "r") as file:
                cdl_string_1 = file.read()

            # ... and now test it as an individual string input
            f_from_str = cf.read(cdl_string_1, cdl_string=True)
            f_from_file = cf.read(tempf)  # len 1 so only one field to check
            self.assertEqual(len(f_from_str), len(f_from_file))
            self.assertEqual(f_from_str[0], f_from_file[0])

            # ... and test further by inputting it in duplicate as a sequence
            f_from_str = cf.read([cdl_string_1, cdl_string_1], cdl_string=True)
            f_from_file = cf.read(tempf)  # len 1 so only one field to check
            self.assertEqual(len(f_from_str), 2 * len(f_from_file))
            self.assertEqual(f_from_str[0], f_from_file[0])
            self.assertEqual(f_from_str[1], f_from_file[0])

            # Check compatibility with the `fmt` kwarg.
            f0 = cf.read(cdl_string_1, cdl_string=True, fmt="CDL")  # fine
            self.assertEqual(len(f0), len(f_from_file))
            self.assertEqual(f0[0], f_from_file[0])
            # If the 'fmt' and 'cdl_string' values contradict each other,
            # alert the user to this. Note that the default fmt is None but
            # it then gets interpreted as NETCDF, so default fmt is fine and
            # it is tested in f_from_str above where fmt is not set.
            with self.assertRaises(ValueError):
                f0 = cf.read(cdl_string_1, cdl_string=True, fmt="NETCDF")

        # If the user forgets the cdl_string=True argument they will
        # accidentally attempt to create a file with a very long name of
        # the CDL string, which will in most, if not all, cases result in
        # an "OSError: [Errno 36] File name too long" error:
        with self.assertRaises(OSError):
            cf.read(cdl_string_1)

    def test_read_write_string(self):
        f = cf.read(self.string_filename)

        n = int(len(f) / 2)

        for i in range(n):
            j = i + n
            self.assertTrue(
                f[i].data.equals(f[j].data, verbose=1), f"{f[i]!r} {f[j]!r}"
            )
            self.assertTrue(
                f[j].data.equals(f[i].data, verbose=1), f"{f[j]!r} {f[i]!r}"
            )

        # Note: Don't loop round all netCDF formats for better
        #       performance. Just one netCDF3 and one netCDF4 format
        #       is sufficient to test the functionality

        for string0 in (True, False):
            for fmt0 in ("NETCDF4", "NETCDF3_CLASSIC"):
                cf.write(f, tmpfile0, fmt=fmt0, string=string0)

                for string1 in (True, False):
                    for fmt1 in ("NETCDF4", "NETCDF3_CLASSIC"):
                        cf.write(f, tmpfile1, fmt=fmt1, string=string1)

                        for i, j in zip(cf.read(tmpfile1), cf.read(tmpfile0)):
                            self.assertTrue(i.equals(j, verbose=1))

    def test_read_broken_bounds(self):
        f = cf.read(self.broken_bounds, verbose=0)
        self.assertEqual(len(f), 2)

    def test_write_coordinates(self):
        f = cf.example_field(0)

        cf.write(f, tmpfile, coordinates=True)
        g = cf.read(tmpfile)

        self.assertEqual(len(g), 1)
        self.assertTrue(g[0].equals(f))

    def test_read_write_domain(self):
        f = cf.read(self.filename)[0]
        d = f.domain

        # 1 domain
        cf.write(d, tmpfile)
        e = cf.read(tmpfile)
        self.assertIsInstance(e, cf.FieldList)
        self.assertTrue(len(e), 10)

        e = cf.read(tmpfile, domain=True, verbose=1)
        self.assertEqual(len(e), 1)
        self.assertIsInstance(e, cf.DomainList)
        e = e[0]
        self.assertIsInstance(e, cf.Domain)
        self.assertTrue(e.equals(e.copy(), verbose=3))
        self.assertTrue(d.equals(e, verbose=3))
        self.assertTrue(e.equals(d, verbose=3))

        # 1 field and 1 domain
        cf.write([f, d], tmpfile)
        g = cf.read(tmpfile)
        self.assertTrue(len(g), 1)
        g = g[0]
        self.assertIsInstance(g, cf.Field)
        self.assertTrue(g.equals(f, verbose=3))

        e = cf.read(tmpfile, domain=True, verbose=1)
        self.assertEqual(len(e), 1)
        e = e[0]
        self.assertIsInstance(e, cf.Domain)

        # 1 field and 2 domains
        cf.write([f, d, d], tmpfile)
        g = cf.read(tmpfile)
        self.assertTrue(len(g), 1)
        g = g[0]
        self.assertIsInstance(g, cf.Field)
        self.assertTrue(g.equals(f, verbose=3))

        e = cf.read(tmpfile, domain=True, verbose=1)
        self.assertEqual(len(e), 2)
        self.assertIsInstance(e[0], cf.Domain)
        self.assertIsInstance(e[1], cf.Domain)
        self.assertTrue(e[0].equals(e[1]))

    def test_read_chunks(self):
        f = cf.example_field(0)
        f.construct("latitude").axis = "Y"
        cf.write(f, tmpfile)

        f = cf.read(tmpfile, chunks={})[0]
        self.assertEqual(f.data.chunks, ((5,), (8,)))

        f = cf.read(tmpfile, chunks=-1)[0]
        self.assertEqual(f.data.chunks, ((5,), (8,)))

        f = cf.read(tmpfile, chunks=None)[0]
        self.assertEqual(f.data.chunks, ((5,), (8,)))

        f = cf.read(tmpfile, chunks={"foo": 2, "bar": 3})[0]
        self.assertEqual(f.data.chunks, ((5,), (8,)))

        with cf.chunksize("200GB"):
            f = cf.read(tmpfile)[0]
            self.assertEqual(f.data.chunks, ((5,), (8,)))

        with cf.chunksize("150B"):
            f = cf.read(tmpfile)[0]
            self.assertEqual(f.data.chunks, ((4, 1), (4, 4)))

        f = cf.read(tmpfile, chunks="150B")[0]
        self.assertEqual(f.data.chunks, ((4, 1), (4, 4)))

        f = cf.read(tmpfile, chunks=3)[0]
        self.assertEqual(f.data.chunks, ((3, 2), (3, 3, 2)))

        y = f.construct("Y")
        self.assertEqual(y.data.chunks, ((3, 2),))

        f = cf.read(tmpfile, chunks={"ncdim%lon": 3})[0]
        self.assertEqual(f.data.chunks, ((5,), (3, 3, 2)))

        f = cf.read(tmpfile, chunks={"longitude": 5, "Y": "150B"})[0]
        self.assertEqual(f.data.chunks, ((3, 2), (5, 3)))

        y = f.construct("Y")
        self.assertEqual(y.data.chunks, ((5,),))

    def test_write_omit_data(self):
        """Test the `omit_data` parameter to `write`."""
        f = cf.example_field(1)
        cf.write(f, tmpfile)

        cf.write(f, tmpfile, omit_data="all")
        g = cf.read(tmpfile)
        self.assertEqual(len(g), 1)
        g = g[0]

        # Check that the data are missing
        self.assertFalse(g.array.count())
        self.assertFalse(g.construct("grid_latitude").array.count())

        # Check that a dump works
        g.dump(display=False)

        cf.write(f, tmpfile, omit_data=("field", "dimension_coordinate"))
        g = cf.read(tmpfile)[0]

        # Check that only the field and dimension coordinate data are
        # missing
        self.assertFalse(g.array.count())
        self.assertFalse(g.construct("grid_latitude").array.count())
        self.assertTrue(g.construct("latitude").array.count())

        cf.write(f, tmpfile, omit_data="field")
        g = cf.read(tmpfile)[0]

        # Check that only the field data are missing
        self.assertFalse(g.array.count())
        self.assertTrue(g.construct("grid_latitude").array.count())


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import dask.array as da


class CompressedArrayMixin:
    """Mixin class for compressed arrays.

    .. versionadded:: TODODASKVER

    """

    def _lock_file_read(self, array):
        """Try to return a dask array that does not support concurrent
        reads.

        .. versionadded:: TODODASKVER

        :Parameters:

            array: array_like
                The array to process.

        :Returns"

            `dask.array.Array` or array_like
                The new `dask` array, or the orginal array if it
                couldn't be ascertained how to form the `dask` array.

        """
        try:
            chunks = array.chunks
        except AttributeError:
            chunks = "auto"

        try:
            array = array.source()
        except (ValueError, AttributeError):
            pass

        try:
            array.get_filename()
        except AttributeError:
            pass
        else:
            array = da.from_array(array, chunks=chunks, lock=True)

        return array



import numpy as np

from ...constants import _stash2standard_name
from ...functions import (
    _DEPRECATION_ERROR_ATTRIBUTE,
    get_subspace,
    load_stash2standard_name,
    parse_indices,
)
from ...umread_lib.umfile import File, Rec
from .abstract import FileArray


class UMArray(FileArray):
    """A sub-array stored in a PP or UM fields file."""

    def __init__(
        self,
        filename=None,
        dtype=None,
        ndim=None,
        shape=None,
        size=None,
        header_offset=None,
        data_offset=None,
        disk_length=None,
        fmt=None,
        word_size=None,
        byte_ordering=None,
        units=False,
        calendar=False,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            filename: `str`
                The file name in normalized, absolute form.

            dtype: `numpy.dtype`
                The data type of the data array on disk.

            shape: `tuple`
                The shape of the unpacked data array. Note that this
                is the shape as required by the object containing the
                `UMArray` object, and so may contain extra size one
                dimensions. When read, the data on disk is reshaped to
                *shape*.

            header_offset: `int`
                The start position in the file of the header.

            data_offset: `int`, optional
                The start position in the file of the data array.

            disk_length: `int`, optional
                The number of words on disk for the data array,
                usually LBLREC-LBEXT. If set to ``0`` then `!size` is
                used.

            fmt: `str`, optional
                ``'PP'`` or ``'FF'``

            word_size: `int`, optional
                ``4`` or ``8``

            byte_ordering: `str`, optional
                ``'little_endian'`` or ``'big_endian'``

            size: `int`
                Deprecated at version TODODASKVER. If set will be
                ignored.

                Number of elements in the uncompressed array.

            ndim: `int`
                Deprecated at version TODODASKVER. If set will be
                ignored.

                The number of uncompressed array dimensions.

            units: `str` or `None`, optional
                The units of the fragment data. Set to `None` to
                indicate that there are no units. If unset then the
                units will be set during the first `__getitem__` call.

            calendar: `str` or `None`, optional
                The calendar of the fragment data. Set to `None` to
                indicate the CF default calendar, if applicable. If
                unset then the calendar will be set during the first
                `__getitem__` call.

            source: optional
                Initialise the array from the given object.

                {{init source}}

            {{deep copy}}

        """
        super().__init__(source=source, copy=copy)

        if source is not None:
            try:
                shape = source._get_component("shape", None)
            except AttributeError:
                shape = None

            try:
                filename = source._get_component("filename", None)
            except AttributeError:
                filename = None

            try:
                fmt = source._get_component("fmt", None)
            except AttributeError:
                fmt = None

            try:
                disk_length = source._get_component("disk_length", None)
            except AttributeError:
                disk_length = None

            try:
                header_offset = source._get_component("header_offset", None)
            except AttributeError:
                header_offset = None

            try:
                data_offset = source._get_component("data_offset", None)
            except AttributeError:
                data_offset = None

            try:
                dtype = source._get_component("dtype", None)
            except AttributeError:
                dtype = None

            try:
                word_size = source._get_component("word_size", None)
            except AttributeError:
                word_size = None

            try:
                byte_ordering = source._get_component("byte_ordering", None)
            except AttributeError:
                byte_ordering = None

            try:
                units = source._get_component("units", False)
            except AttributeError:
                units = False

            try:
                calendar = source._get_component("calendar", False)
            except AttributeError:
                calendar = False

        self._set_component("shape", shape, copy=False)
        self._set_component("filename", filename, copy=False)
        self._set_component("dtype", dtype, copy=False)
        self._set_component("header_offset", header_offset, copy=False)
        self._set_component("data_offset", data_offset, copy=False)
        self._set_component("disk_length", disk_length, copy=False)
        self._set_component("units", units, copy=False)
        self._set_component("calendar", calendar, copy=False)

        if fmt is not None:
            self._set_component("fmt", fmt, copy=False)

        if byte_ordering is not None:
            self._set_component("byte_ordering", byte_ordering, copy=False)

        if word_size is not None:
            self._set_component("word_size", word_size, copy=False)

        # By default, close the UM file after data array access
        self._set_component("close", True, copy=False)

    def __getitem__(self, indices):
        """Return a subspace of the array.

        x.__getitem__(indices) <==> x[indices]

        Returns a subspace of the array as an independent numpy array.

        """
        f = self.open()
        rec = self._get_rec(f)

        int_hdr = rec.int_hdr
        real_hdr = rec.real_hdr
        array = rec.get_data().reshape(self.shape)

        self.close(f)
        del f

        if indices is not Ellipsis:
            indices = parse_indices(array.shape, indices)
            array = get_subspace(array, indices)

        # Set the units, if they haven't been set already.
        self._set_units(int_hdr)

        LBUSER2 = int_hdr.item(38)
        if LBUSER2 == 3:
            # Return the numpy array now if it is a boolean array
            self._set_component("dtype", np.dtype(bool), copy=False)
            return array.astype(bool)

        integer_array = LBUSER2 == 2

        # ------------------------------------------------------------
        # Convert to a masked array
        # ------------------------------------------------------------
        # Set the fill_value from BMDI
        fill_value = real_hdr.item(17)
        if fill_value != -1.0e30:
            # -1.0e30 is the flag for no missing data
            if integer_array:
                # The fill_value must be of the same type as the data
                # values
                fill_value = int(fill_value)

            # Mask any missing values
            mask = array == fill_value
            if mask.any():
                array = np.ma.masked_where(mask, array, copy=False)

        # ------------------------------------------------------------
        # Unpack the array using the scale_factor and add_offset, if
        # either is available
        # ------------------------------------------------------------
        # Treat BMKS as a scale_factor if it is neither 0 nor 1
        scale_factor = real_hdr.item(18)
        if scale_factor != 1.0 and scale_factor != 0.0:
            if integer_array:
                scale_factor = int(scale_factor)

            array *= scale_factor

        # Treat BDATUM as an add_offset if it is not 0
        add_offset = real_hdr.item(4)
        if add_offset != 0.0:
            if integer_array:
                add_offset = int(add_offset)

            array += add_offset

        # Set the data type
        self._set_component("dtype", array.dtype, copy=False)

        # Return the numpy array
        return array

    def _get_rec(self, f):
        """Get a container for a record.

        This includes the lookup header and file offsets.

        .. versionadded:: TODODASKVER

        .. seealso:: `close`, `open`

        :Parameters:

            f: `umread_lib.umfile.File`
                The open PP or FF file.

        :Returns:

            `umread_lib.umfile.Rec`
                The record container.

        """
        header_offset = self.header_offset
        data_offset = self.data_offset
        disk_length = self.disk_length
        if data_offset is None or disk_length is None:
            # This method doesn't require data_offset and disk_length,
            # so plays nicely with CFA. Is it fast enough that we can
            # use this method always?
            for v in f.vars:
                for r in v.recs:
                    if r.hdr_offset == header_offset:
                        return r
        else:
            return Rec.from_file_and_offsets(
                f, header_offset, data_offset, disk_length
            )

    def _set_units(self, int_hdr):
        """The units and calendar properties.

        These are set from inpection of the integer header, but only
        if they have already not been defined, either during {{class}}
        instantiation or by a previous call to `_set_units`.

        .. versionadded:: TODODASKVER

        :Parameters:

            int_hdr: `numpy.ndarray`
                The integer header of the data.

        :Returns:

            `tuple`
                The units and calendar values, either of which may be
                `None`.

        """
        units = self._get_component("units", False)
        if units is False:
            units = None

            if not _stash2standard_name:
                load_stash2standard_name()

            submodel = int_hdr[44]
            stash = int_hdr[41]
            records = _stash2standard_name.get((submodel, stash))
            if records:
                LBSRCE = int_hdr[37]
                version, source = divmod(LBSRCE, 10000)
                if version <= 0:
                    version = 405.0

                for (
                    long_name,
                    units0,
                    valid_from,
                    valid_to,
                    standard_name,
                    cf_info,
                    condition,
                ) in records:
                    if not self._test_version(
                        valid_from, valid_to, version
                    ) or not self._test_condition(condition, int_hdr):
                        continue

                    units = units0
                    break

            self._set_component("units", units, copy=False)

        calendar = self._get_component("calendar", False)
        if calendar is False:
            calendar = None
            self._set_component("calendar", calendar, copy=False)

        return units, calendar

    def _test_condition(self, condition, int_hdr):
        """Return `True` if a field satisfies a condition for a STASH
        code to standard name conversion.

        .. versionadded:: TODODASKVER

        :Parameters:

            condition: `str`
                The condition. If False then the condition is always
                passed, otherwise the condition is specified as
                ``'true_latitude_longitude'`` or
                ``'rotated_latitude_longitude'``.

            int_hdr: `numpy.ndarray`
                The integer lookup header used to evaluate the
                condition.

        :Returns:

            `bool`
                `True` if the data satisfies the condition specified,
                `False` otherwise.

        """
        if not condition:
            return True

        if condition == "true_latitude_longitude":
            LBCODE = int_hdr[15]
            # LBCODE 1: Unrotated regular lat/long grid
            # LBCODE 2 = Regular lat/lon grid boxes (grid points are
            #            box centres)
            if LBCODE in (1, 2):
                return True
        elif condition == "rotated_latitude_longitude":
            LBCODE = int_hdr[15]
            # LBCODE 101: Rotated regular lat/long grid
            # LBCODE 102: Rotated regular lat/lon grid boxes (grid
            #             points are box centres)
            # LBCODE 111: ?
            if LBCODE in (101, 102, 111):
                return True
        else:
            return False

    def _test_version(self, valid_from, valid_to, version):
        """Return `True` if the UM version applicable to this field is
        within the given range.

        If possible, the UM version is derived from the PP header and
        stored in the metadata object. Otherwise it is taken from the
        *version* parameter.

        .. versionadded:: TODODASKVER

        :Parameters:

            valid_from: number or `None`
                The lower bound of the version range, e.g. ``4.5``,
                ``606.1``, etc.

            valid_to: number or `None`
                The upper bound of the version range, e.g. ``4.5``,
                ``606.1``, etc.

            version: number
                The version of field, e.g. ``4.5``, ``606.1``, etc.

        :Returns:

            `bool`
                `True` if the UM version applicable to this data is
                within the given range, `False` otherwise.

        """
        if valid_to is None:
            if valid_from is None:
                return True

            if valid_from <= version:
                return True
        elif valid_from is None:
            if version <= valid_to:
                return True
        elif valid_from <= version <= valid_to:
            return True

        return False

    @property
    def file_address(self):
        """The file name and address.

        Deprecated at version TODODASKVER. Use methods `get_filename`
        and `get_address` instead.

        :Returns:

            `tuple`
                The file name and file address.

        **Examples**

        >>> a.file_address()
        ('file.pp', 234835)

        """
        _DEPRECATION_ERROR_ATTRIBUTE(
            self,
            "file_address",
            "Use methods 'get_filename' and 'get_address' instead.",
            version="TODODASKVER",
            removed_at="5.0.0",
        )  # pragma: no cover

    @property
    def header_offset(self):
        """The start position in the file of the header.

        :Returns:

            `int`

        """
        return self._get_component("header_offset")

    @property
    def data_offset(self):
        """The start position in the file of the data array.

        :Returns:

            `int`

        """
        return self._get_component("data_offset")

    @property
    def disk_length(self):
        """The number of words on disk for the data array.

        :Returns:

            `int`

        """
        return self._get_component("disk_length")

    @property
    def fmt(self):
        """The file format of the UM file containing the array.

        Deprecated at version TODODASKVER. Use method `get_fmt`
        instead.

        :Returns:

            `str`
                'FF' or 'PP'

        """
        _DEPRECATION_ERROR_ATTRIBUTE(
            self,
            "fmt",
            "Use method 'get_fmt' instead.",
            version="TODODASKVER",
            removed_at="5.0.0",
        )  # pragma: no cover

    @property
    def byte_ordering(self):
        """The endianness of the data.

        Deprecated at version TODODASKVER. Use method
        `get_byte_ordering` instead.

        :Returns:

            `str`
                'little_endian' or 'big_endian'

        """
        _DEPRECATION_ERROR_ATTRIBUTE(
            self,
            "byte_ordering",
            "Use method 'get_byte_ordering' instead.",
            version="TODODASKVER",
            removed_at="5.0.0",
        )  # pragma: no cover

    @property
    def word_size(self):
        """Word size in bytes.

        Deprecated at version TODODASKVER. Use method `get_word_size`
        instead.

        :Returns:

            `int`
                4 or 8

        """
        _DEPRECATION_ERROR_ATTRIBUTE(
            self,
            "word_size",
            "Use method 'get_word_size' instead.",
            version="TODODASKVER",
            removed_at="5.0.0",
        )  # pragma: no cover

    def close(self, f):
        """Close the dataset containing the data.

        :Parameters:

            f: `umfile_lib.File`
                The UM or PP dataset to be be closed.

                .. versionadded:: TODODASKVER

        :Returns:

            `None`

        """
        if self._get_component("close"):
            f.close_fd()

    def get_address(self):
        """The address in the file of the variable.

        The address is the word offset of the lookup header.

        .. versionadded:: TODODASKVER

        :Returns:

            `str` or `None`
                The address, or `None` if there isn't one.

        """
        return self.header_offset

    def get_byte_ordering(self):
        """The endianness of the data.

        .. versionadded:: TODODASKVER

        .. seealso:: `open`

        :Returns:

            `str` or `None`
                ``'little_endian'`` or ``'big_endian'``. If the byte
                ordereing has not been set then `None` is returned, in
                which case byte ordering will be detected
                automatically (if possible) when the file is opened
                with `open`.

        """
        return self._get_component("byte_ordering", None)

    def get_fmt(self):
        """The file format of the UM file containing the array.

        .. versionadded:: TODODASKVER

        .. seealso:: `open`

        :Returns:

            `str` or `None`
                ``'FF'`` or ``'PP'``. If the word size has not been
                set then `None` is returned, in which case file format
                will be detected automatically (if possible) when the
                file is opened with `open`.

        """
        return self._get_component("fmt", None)

    def get_word_size(self):
        """Word size in bytes.

        .. versionadded:: TODODASKVER

        .. seealso:: `open`

        :Returns:

            `int` or `None`
                ``4`` or ``8``. If the word size has not been set then
                `None` is returned, in which case word size will be
                detected automatically (if possible) when the file is
                opened with `open`.

        """
        return self._get_component("word_size", None)

    def open(self):
        """Returns an open dataset containing the data array.

        :Returns:

            `umfile_lib.File`

        **Examples**

        >>> f.open()

        """
        try:
            f = File(
                path=self.get_filename(),
                byte_ordering=self.get_byte_ordering(),
                word_size=self.get_word_size(),
                fmt=self.get_fmt(),
            )
        except Exception as error:
            try:
                f.close_fd()
            except Exception:
                pass

            raise Exception(error)
        else:
            return f
